﻿using FliUtils.Mvvm;
using FliProCameraLib;
using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.IO;
using System.Xml.Serialization;

namespace FLIPilotSimpleViewModel
{
    public partial class ControlPanelViewModel : ObservableProperties
    {

        //**************************************************************
        #region Constructor and Initialization

        private ControlPanelViewModel()
        {
        }

        #endregion

        //**************************************************************
        #region Constants

        public const int IMAGE_HEIGHT_MAX = 2224;
        public const int IMAGE_WIDTH_MAX = 2840;

        public const int WINDOW_ONE_INDEX = 1;
        public const int WINDOW_THREE_INDEX = 2;

        public const int TAB_INDEX_GRAB_PANEL = 0;
        public const int TAB_INDEX_EXPOSURE_PANEL = 1;
        public const int TAB_INDEX_FRAME_PANEL = 2;
        public const int TAB_INDEX_SENSOR_PANEL = 3;
        public const int TAB_INDEX_TRIGGER_PANEL = 4;
        public const int TAB_INDEX_CAMERA_PANEL = 5;
        public const int TAB_INDEX_TRACKING_PANEL = 6;
        public const int TAB_INDEX_MERGE_PANEL = 7;
        public const int TAB_INDEX_FILE_PANEL = 8;

        #endregion

        //**************************************************************
        #region Public Types

        public class ContextMenuItem
        {
            public ContextMenuItem() { }
            public ContextMenuItem(string header, int tabIndex)
            {
                Header = header;
                TabIndex = tabIndex;
            }
            public string Header { get; set; }
            public int TabIndex { get; set; }
        }

        public class ImageActionEventArgs : EventArgs
        {
            public ImageActionEnum Action { get; set; }
            public object Value { get; set; }
        }

        public enum ImageActionEnum
        {
            RotateImage,
            FlipLeftToRight,
            FlipTopToBottom,
            ResetImage
        }

        public enum ImageAnalysisEnum
        {
            None,
            ColumnAndRowProfile,
            RegionProfile,
            SingleLineProfile,
            ColumnAndRowHistogram,
            RegionHistogram,
            SingleLineHistogram
        }

        #endregion

        //**************************************************************
        #region Protected Properties

        #endregion Protected Properties

        //**************************************************************
        #region Observable Properties

        //-------------------------------------------
        /// <summary>
        /// The Active Electrons Per Count Low to set default Merge settings
        /// </summary>
        public double ActiveElectronsPerCountHigh
        {
            get { return _ActiveElectronsPerCountHigh; }
            set
            {
                if (SetValue(ref _ActiveElectronsPerCountHigh, value))
                {
                }
            }
        }
        private double _ActiveElectronsPerCountHigh;

        //-------------------------------------------
        /// <summary>
        /// The Active Electrons Per Count Low to set default Merge settings
        /// </summary>
        public double ActiveElectronsPerCountLow
        {
            get { return _ActiveElectronsPerCountLow; }
            set { SetValue(ref _ActiveElectronsPerCountLow, value); }
        }
        private double _ActiveElectronsPerCountLow;

        //-------------------------------------------
        /// <summary>
        /// The ADUs per electron  ratio
        /// </summary>
        public double ADUsPerElectron
        {
            get { return _ADUsPerElectron; }
            set { SetValue(ref _ADUsPerElectron, value); }
        }
        private double _ADUsPerElectron = 1;

        //-------------------------------------------
        /// <summary>
        /// The black level offset of the High Gain Image (OH - DN)
        /// </summary>
        public int BlackOffsetH
        {
            get { return _BlackOffsetH; }
            set { SetValue(ref _BlackOffsetH, value); }
        }
        private int _BlackOffsetH = 154;

        //-------------------------------------------
        /// <summary>
        /// The black level offset of the Low Gain Image (OL - DN)
        /// </summary>
        public int BlackOffsetL
        {
            get { return _BlackOffsetL; }
            set { SetValue(ref _BlackOffsetL, value); }
        }
        private int _BlackOffsetL = 129;
        
        //-------------------------------------------
        /// <summary>
        /// The Gain Ratio
        /// </summary>
        public double GainRatio
        {
            get { return _GainRatio; }
            set { SetValue(ref _GainRatio, value); }
        }
        private double _GainRatio = 44.94;

        //-------------------------------------------
        /// <summary>
        /// The parameter TH (DN) is usually set to 3800 DN which is the maximum value of 
        /// the linear region of the response curve for high gain image.
        /// </summary>
        public int HighLowThreshold
        {
            get { return _HighLowThreshold; }
            set { SetValue(ref _HighLowThreshold, value); }
        }
        private int _HighLowThreshold = 3800;

        public ImageAnalysisEnum ImageAnalysisValue
        {
            get { return _ImageAnalysisValue; }
            set
            {
                if (SetValue(ref _ImageAnalysisValue, value))
                {
                    OnHistogramDataAction(new EventArgs());
                }
            }
        }
        private ImageAnalysisEnum _ImageAnalysisValue = ImageAnalysisEnum.None;

        public bool ImageAnalysisNone
        {
            get { return ImageAnalysisValue == ImageAnalysisEnum.None; }
            set
            {
                if (value)
                {
                    ImageAnalysisValue = ImageAnalysisEnum.None;
                }
            }
        }
        
        public bool ImageAnalysisScaleData
        {
            get { return _ImageAnalysisScaleData; }
            set { SetValue(ref _ImageAnalysisScaleData, value); }
        }
        private bool _ImageAnalysisScaleData = false;

        public bool ImageAnalysisColumnAndRowHistogram
        {
            get { return ImageAnalysisValue == ImageAnalysisEnum.ColumnAndRowHistogram; }
            set
            {
                if (value)
                {
                    ImageAnalysisValue = ImageAnalysisEnum.ColumnAndRowHistogram;
                }
            }
        }

        public bool ImageAnalysisColumnAndRowProfile
        {
            get { return ImageAnalysisValue == ImageAnalysisEnum.ColumnAndRowProfile; }
            set
            {
                if (value)
                {
                    ImageAnalysisValue = ImageAnalysisEnum.ColumnAndRowProfile;
                }
            }
        }

        public bool ImageAnalysisRegionHistogram
        {
            get { return ImageAnalysisValue == ImageAnalysisEnum.RegionHistogram; }
            set
            {
                if (value)
                {
                    ImageAnalysisValue = ImageAnalysisEnum.RegionHistogram;
                }
            }
        }

        public bool ImageAnalysisRegionProfile
        {
            get { return ImageAnalysisValue == ImageAnalysisEnum.RegionProfile; }
            set
            {
                if (value)
                {
                    ImageAnalysisValue = ImageAnalysisEnum.RegionProfile;
                }
            }
        }

        public bool ImageAnalysisSingleLineHistogram
        {
            get { return ImageAnalysisValue == ImageAnalysisEnum.SingleLineHistogram; }
            set
            {
                if (value)
                {
                    ImageAnalysisValue = ImageAnalysisEnum.SingleLineHistogram;
                }
            }
        }

        public bool ImageAnalysisSingleLineProfile
        {
            get { return ImageAnalysisValue == ImageAnalysisEnum.SingleLineProfile; }
            set
            {
                if (value)
                {
                    ImageAnalysisValue = ImageAnalysisEnum.SingleLineProfile;
                }
            }
        }

        /// <summary>
        /// The offset between the HDR slope and the LDR slope
        /// </summary>
        public int LineOffset
        {
            get { return _LineOffset; }
            set { SetValue(ref _LineOffset, value); }
        }
        private int _LineOffset = 8000;

        /// <summary>
        /// The offset at the first temperature Merge settings between the HDR slope and the LDR slope
        /// </summary>
        public int LineOffset1
        {
            get { return _LineOffset1; }
            set { SetValue(ref _LineOffset1, value); }
        }
        private int _LineOffset1 = 1700;

        /// <summary>
        /// The offset at the second temperature Merge settings between the HDR slope and the LDR slope
        /// </summary>
        public int LineOffset2
        {
            get { return _LineOffset2; }
            set { SetValue(ref _LineOffset2, value); }
        }
        private int _LineOffset2 = 1750;

        /// <summary>
        /// The offset at the third temperature Merge settings between the HDR slope and the LDR slope
        /// </summary>
        public int LineOffset3
        {
            get { return _LineOffset3; }
            set { SetValue(ref _LineOffset3, value); }
        }
        private int _LineOffset3 = 1800;

        public int MergeRange1Count
        {
            get { return _MergeRange1Count; }
            set { SetValue<int>(ref _MergeRange1Count, value); }
        }
        protected int _MergeRange1Count;

        public int MergeRange2Count
        {
            get { return _MergeRange2Count; }
            set { SetValue<int>(ref _MergeRange2Count, value); }
        }
        protected int _MergeRange2Count;

        public int MergeRange3Count
        {
            get { return _MergeRange3Count; }
            set
            {
                SetValue<int>(ref _MergeRange3Count, value);
            }
        }
        protected int _MergeRange3Count;

        public int MergeRange1Lower
        {
            get { return _MergeRange1Lower; }
            set { SetValue<int>(ref _MergeRange1Lower, value); }
        }
        protected int _MergeRange1Lower;

        public int MergeRange2Lower
        {
            get { return _MergeRange2Lower; }
            set { SetValue<int>(ref _MergeRange2Lower, value); }
        }
        protected int _MergeRange2Lower;

        public int MergeRange3Lower
        {
            get { return _MergeRange3Lower; }
            set { SetValue<int>(ref _MergeRange3Lower, value); }
        }
        protected int _MergeRange3Lower;

        public int MergeRange1Upper
        {
            get { return _MergeRange1Upper; }
            set { SetValue<int>(ref _MergeRange1Upper, value); }
        }
        protected int _MergeRange1Upper;

        public int MergeRange2Upper
        {
            get { return _MergeRange2Upper; }
            set { SetValue<int>(ref _MergeRange2Upper, value); }
        }
        protected int _MergeRange2Upper;

        public int MergeRange3Upper
        {
            get { return _MergeRange3Upper; }
            set { SetValue<int>(ref _MergeRange3Upper, value); }
        }
        protected int _MergeRange3Upper;

        public string MergeRange1Text
        {
            get
            {
                return "(" + MergeRange1Lower.ToString() + " - " + MergeRange1Upper.ToString() + ")";
            }
        }

        public string MergeRange2Text
        {
            get
            {
                return "(" + MergeRange2Lower.ToString() + " - " + MergeRange2Upper.ToString() + ")";
            }
        }

        public string MergeRange3Text
        {
            get
            {
                return "(" + MergeRange3Lower.ToString() + " - " + MergeRange3Upper.ToString() + ")";
            }
        }
        
        //-------------------------------------------
        /// <summary>
        /// The First Temperature to set default Merge settings
        /// </summary>
        public double Temperature1
        {
            get { return _Temperature1; }
            set { SetValue(ref _Temperature1, value); }
        }
        private double _Temperature1 = 20;

        //-------------------------------------------
        /// <summary>
        /// The Second Temperature to set default Merge settings
        /// </summary>
        public double Temperature2
        {
            get { return _Temperature2; }
            set { SetValue(ref _Temperature2, value); }
        }
        private double _Temperature2 = 5;

        //-------------------------------------------
        /// <summary>
        /// The Third Temperature to set default Merge settings
        /// </summary>
        public double Temperature3
        {
            get { return _Temperature3; }
            set { SetValue(ref _Temperature3, value); }
        }
        private double _Temperature3 = -10;

        #endregion

        //**************************************************************
        #region Public Properties

        //-------------------------------------------
        /// <summary>
        /// Returns a reference to the single instance of this class.
        /// </summary>
        public static ControlPanelViewModel Instance
        {
            get
            {
                if (_instance == null)
                {
                    lock (syncRoot)
                    {
                        if (_instance == null)
                        {
                            _instance = new ControlPanelViewModel();
                        }
                    }
                }
                return _instance;
            }
        }

        #endregion Public Properties

        //**************************************************************
        #region Public Classes

        /// <summary>
        /// This class provides the structure to register UI Tabs with an index
        /// </summary>
        public class TabAction
        {
            public TabAction()
            {
            }

            public TabAction(int index, TabItem tab, Expander expanderItem)
            {
                Index = index;
                Tab = tab;
                ExpanderItem = expanderItem;
            }

            public int Index { get; set; }
            public TabItem Tab { get; set; }
            public Expander ExpanderItem { get; set; }
        }

        #endregion

        //**************************************************************
        #region Public Lists

        public List<ContextMenuItem> ContextMenuList = new List<ContextMenuItem>();

        #endregion

        //**************************************************************
        #region Event Handlers

        /// <summary>
        /// Sends the Abort command to the camera
        /// </summary>
        public void Abort()
        {
            MessageBox.Show("Abort called", "Warning", MessageBoxButton.OK,
                MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
        }

        /// <summary>
        /// Sends the FlipLeftToRight command to the MainWindow to be sent to the images
        /// </summary>
        public void FlipLeftToRight(bool isChecked)
        {
            ImageActionEventArgs args = new ImageActionEventArgs() { Action = ImageActionEnum.FlipLeftToRight, Value = isChecked };
            OnImageAction(args);
        }

        /// <summary>
        /// Sends the FlipTopToBottom command to the MainWindow to be sent to the images
        /// </summary>
        public void FlipTopToBottom(bool isChecked)
        {
            ImageActionEventArgs args = new ImageActionEventArgs() { Action = ImageActionEnum.FlipTopToBottom, Value = isChecked };
            OnImageAction(args);
        }

        /// <summary>
        /// Sends the selected Grab command to the camera
        /// </summary>
        public void Grab()
        {
            if (fliProCameraManager.ActiveIndex < 0)
            {
                MessageBox.Show("There are no Cameras connected.", "Warning", MessageBoxButton.OK,
                    MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
            }
            else if (fliProCameraManager.AvailableCameras[fliProCameraManager.ActiveIndex].IsFromFile)
            {
                MessageBox.Show("Grab is not valid for Files. Select a Camera from the drop list.", "Warning", MessageBoxButton.OK,
                    MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
            }
            else if (fliProCameraManager.AvailableCameras[fliProCameraManager.ActiveIndex].IsGeneratedBitmap)
            {
                MessageBox.Show("Grab is not valid for Generated Bitmaps. Select a Camera from the drop list.", "Warning", MessageBoxButton.OK,
                    MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
            }
            else if (!fliProCameraManager.ActiveCamera.IsConnected)
            {
                MessageBox.Show("Camera " + fliProCameraManager.ActiveCamera.SerialNumber + 
                    " is no longer connected.\nSelect a Camera from the drop list that is connected.", "Warning", MessageBoxButton.OK,
                    MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
            }
            else
            {
                fliProCameraManager.ActiveCamera.GrabFrame();
            }
        }

        #endregion

        //**************************************************************
        #region Public Functions

        public void ApplyHistogramSettings()
        {
            OnHistogramSettingsAction(new EventArgs());
        }

        public void ApplyMergeSettings()
        {
            fliProCameraManager.DisplayCamera.ADUsPerElectron = ADUsPerElectron;
            fliProCameraManager.DisplayCamera.ADUsPerElectronChangeAction += NotifyADUsPerElectronChangeActionProc;
            fliProCameraManager.DisplayCamera.GainRatio = GainRatio;
            fliProCameraManager.DisplayCamera.LineOffset = LineOffset;
            fliProCameraManager.DisplayCamera.HighLowThreshold = HighLowThreshold;
            fliProCameraManager.DisplayCamera.Temperature1 = Temperature1;
            fliProCameraManager.DisplayCamera.Temperature2 = Temperature2;
            fliProCameraManager.DisplayCamera.Temperature3 = Temperature3;
            fliProCameraManager.DisplayCamera.LineOffset1 = LineOffset1;
            fliProCameraManager.DisplayCamera.LineOffset2 = LineOffset2;
            fliProCameraManager.DisplayCamera.LineOffset3 = LineOffset3;
            fliProCameraManager.DisplayCamera.ActiveElectronsPerCountLow = ActiveElectronsPerCountLow;
            fliProCameraManager.DisplayCamera.ActiveElectronsPerCountHigh = ActiveElectronsPerCountHigh;
            fliProCameraManager.DisplayCamera.GainRatioChangeAction += NotifyGainRatioChangeActionProc;
            fliProCameraManager.DisplayCamera.MergeRangeChangeAction += NotifyMergeRangeChangeActionProc;
            fliProCameraManager.DisplayCamera.LineOffsetChangeAction += NotifyLineOffsetChangeActionProc;
        }

        public void ApplyNotifyActionProcsToCamera()
        {
            fliProCameraManager.DisplayCamera.ADUsPerElectronChangeAction += NotifyADUsPerElectronChangeActionProc;
            fliProCameraManager.DisplayCamera.GainRatioChangeAction += NotifyGainRatioChangeActionProc;
            fliProCameraManager.DisplayCamera.MergeRangeChangeAction += NotifyMergeRangeChangeActionProc;
            fliProCameraManager.DisplayCamera.LineOffsetChangeAction += NotifyLineOffsetChangeActionProc;
        }
        
        public void ClearAllMergeSettings()
        {
            fliProCameraManager.ActiveCamera.ClearAllMergeSettings();
        }

        public void DeleteMergeSettings()
        {
            fliProCameraManager.ActiveCamera.DeleteMergeSettings(fliProCameraManager.ActiveCamera.HighGainIndex, fliProCameraManager.ActiveCamera.LowGainIndex);
        }

        public GrabPanelViewModel GetGrabPanelViewModel()
        {
            return GPVM;
        }

        public void NewCameraAction()
        {
            if (!fliProCameraManager.DisplayCamera.MergeValuesChangeActionRegistered)
            {
                fliProCameraManager.DisplayCamera.MergeValuesChangeActionRegistered = true;
                fliProCameraManager.DisplayCamera.MergeValuesChangeAction += NotifyMergeValuesChangeActionProc;
            }
            PopulateFactorySettings();
        }

        public void NotifyADUsPerElectronChangeActionProc(object sender, EventArgs e)
        {
            ADUsPerElectron = fliProCameraManager.DisplayCamera.ADUsPerElectron;
        }

        public void NotifyGainRatioChangeActionProc(object sender, EventArgs e)
        {
            GainRatio = Math.Round(fliProCameraManager.DisplayCamera.GainRatio, 2);
        }

        public void NotifyMergeRangeChangeActionProc(object sender, EventArgs e)
        {
            MergeRange1Count = fliProCameraManager.DisplayCamera.MergeRange1Count;
            MergeRange2Count = fliProCameraManager.DisplayCamera.MergeRange2Count;
            MergeRange3Count = fliProCameraManager.DisplayCamera.MergeRange3Count;
            MergeRange1Lower = fliProCameraManager.DisplayCamera.MergeRange1Lower;
            MergeRange2Lower = fliProCameraManager.DisplayCamera.MergeRange2Lower;
            MergeRange3Lower = fliProCameraManager.DisplayCamera.MergeRange3Lower;
            MergeRange1Upper = fliProCameraManager.DisplayCamera.MergeRange1Upper;
            MergeRange2Upper = fliProCameraManager.DisplayCamera.MergeRange2Upper;
            MergeRange3Upper = fliProCameraManager.DisplayCamera.MergeRange3Upper;
            OnPropertyChanged("MergeRange1Text");
            OnPropertyChanged("MergeRange2Text");
            OnPropertyChanged("MergeRange3Text");
        }

        private void NotifyMergeValuesChangeActionProc(object sender, EventArgs e)
        {
            PopulateFactorySettings();
        }

        public void NotifyLineOffsetChangeActionProc(object sender, EventArgs e)
        {
            LineOffset = fliProCameraManager.DisplayCamera.LineOffset;
        }

        public void PopulateWithToolSettings()
        {
        }

        public void PopulateFactorySettings()
        {
            ADUsPerElectron = fliProCameraManager.ActiveCamera.ADUsPerElectron;
            GainRatio = Math.Round(fliProCameraManager.ActiveCamera.GainRatio, 2);

            // For now we are only coding for the 4040
            Temperature1 = fliProCameraManager.ActiveCamera.Temperature1;
            Temperature2 = fliProCameraManager.ActiveCamera.Temperature2;
            Temperature3 = fliProCameraManager.ActiveCamera.Temperature3;
            LineOffset1 = fliProCameraManager.ActiveCamera.LineOffset1;
            LineOffset2 = fliProCameraManager.ActiveCamera.LineOffset2;
            LineOffset3 = fliProCameraManager.ActiveCamera.LineOffset3;
            ActiveElectronsPerCountHigh = fliProCameraManager.ActiveCamera.ActiveElectronsPerCountHigh;
            ActiveElectronsPerCountLow = fliProCameraManager.ActiveCamera.ActiveElectronsPerCountLow;
        }

        public void ReadDefaultMergeSettings()
        {
            // Now find the target values
            int index = fliProCameraManager.ActiveCamera.GetMergeLineOffsetIndex(fliProCameraManager.ActiveCamera.HighGainIndex, fliProCameraManager.ActiveCamera.LowGainIndex);
            
            if (index < 0)
            {
                // try getting the factory settings
                index = fliProCameraManager.ActiveCamera.GetFactoryMergeLineOffsetIndex(fliProCameraManager.ActiveCamera.HighGainIndex, fliProCameraManager.ActiveCamera.LowGainIndex);
            }
            
            // Set the display to whatever the camera values are
            ADUsPerElectron = fliProCameraManager.ActiveCamera.ADUsPerElectron;
            GainRatio = Math.Round(fliProCameraManager.ActiveCamera.GainRatio, 2);
            Temperature1 = fliProCameraManager.ActiveCamera.Temperature1;
            Temperature2 = fliProCameraManager.ActiveCamera.Temperature2;
            Temperature3 = fliProCameraManager.ActiveCamera.Temperature3;
            LineOffset1 = fliProCameraManager.ActiveCamera.LineOffset1;
            LineOffset2 = fliProCameraManager.ActiveCamera.LineOffset2;
            LineOffset3 = fliProCameraManager.ActiveCamera.LineOffset3;
            ActiveElectronsPerCountHigh = fliProCameraManager.ActiveCamera.ActiveElectronsPerCountHigh;
            ActiveElectronsPerCountLow = fliProCameraManager.ActiveCamera.ActiveElectronsPerCountLow;
        }

        public void RestoreMergeSettingsFromFactorDefaults()
        {
            int index = fliProCameraManager.ActiveCamera.GetFactoryMergeLineOffsetIndex(fliProCameraManager.ActiveCamera.HighGainIndex, fliProCameraManager.ActiveCamera.LowGainIndex);

            ADUsPerElectron = fliProCameraManager.ActiveCamera.ADUsPerElectron;
            GainRatio = Math.Round(fliProCameraManager.ActiveCamera.GainRatio, 2);

            // For now we are only coding for the 4040
            Temperature1 = fliProCameraManager.ActiveCamera.Temperature1;
            Temperature2 = fliProCameraManager.ActiveCamera.Temperature2;
            Temperature3 = fliProCameraManager.ActiveCamera.Temperature3;
            LineOffset1 = fliProCameraManager.ActiveCamera.LineOffset1;
            LineOffset2 = fliProCameraManager.ActiveCamera.LineOffset2;
            LineOffset3 = fliProCameraManager.ActiveCamera.LineOffset3;
            ActiveElectronsPerCountHigh = fliProCameraManager.ActiveCamera.ActiveElectronsPerCountHigh;
            ActiveElectronsPerCountLow = fliProCameraManager.ActiveCamera.ActiveElectronsPerCountLow;
        }

        public void SaveAsDefaultMergeSettings()
        {
            // Copy control panel settings to the camera
            fliProCameraManager.ActiveCamera.ADUsPerElectron = ADUsPerElectron;
            fliProCameraManager.ActiveCamera.GainRatio = GainRatio;

            // For now we are only coding for the 4040
            fliProCameraManager.ActiveCamera.Temperature1 = Temperature1;
            fliProCameraManager.ActiveCamera.Temperature2 = Temperature2;
            fliProCameraManager.ActiveCamera.Temperature3 = Temperature3;
            fliProCameraManager.ActiveCamera.LineOffset1 = LineOffset1;
            fliProCameraManager.ActiveCamera.LineOffset2 = LineOffset2;
            fliProCameraManager.ActiveCamera.LineOffset3 = LineOffset3;
            ActiveElectronsPerCountHigh = fliProCameraManager.ActiveCamera.ActiveElectronsPerCountHigh;
            ActiveElectronsPerCountLow = fliProCameraManager.ActiveCamera.ActiveElectronsPerCountLow;

            fliProCameraManager.ActiveCamera.StoreNVDataToCamera();
        }

        public void SetTab(int index)
        {
            TabItem tab = null;
            Expander expanderItem = null;
            for (int i = 0; i < TabList.Count; i++)
            {
                if (TabList[i].Index == index)
                {
                    tab = TabList[i].Tab;
                    expanderItem = TabList[i].ExpanderItem;
                    break;
                }
            }

            if (expanderItem != null)
            {
                expanderItem.IsExpanded = true;   // Make sure the expander is open
                if (tab != null)
                {
                    tab.IsSelected = true;
                }
            }
        }

        public void TabListAdd(int index, TabItem tab, Expander expanderItem)
        {
            TabAction tabAction = new TabAction(index, tab, expanderItem);
            TabList.Add(tabAction);
        }

        public void UpdateMergeSettings()
        {
            ADUsPerElectron = fliProCameraManager.DisplayCamera.ADUsPerElectron;
            GainRatio = fliProCameraManager.DisplayCamera.GainRatio;
            LineOffset = fliProCameraManager.DisplayCamera.LineOffset;
            Temperature1 = fliProCameraManager.DisplayCamera.Temperature1;
            Temperature2 = fliProCameraManager.DisplayCamera.Temperature2;
            Temperature3 = fliProCameraManager.DisplayCamera.Temperature3;
            LineOffset1 = fliProCameraManager.DisplayCamera.LineOffset1;
            LineOffset2 = fliProCameraManager.DisplayCamera.LineOffset2;
            LineOffset3 = fliProCameraManager.DisplayCamera.LineOffset3;
            ActiveElectronsPerCountLow = fliProCameraManager.DisplayCamera.ActiveElectronsPerCountLow;
            ActiveElectronsPerCountHigh = fliProCameraManager.DisplayCamera.ActiveElectronsPerCountHigh;
            HighLowThreshold = fliProCameraManager.DisplayCamera.HighLowThreshold;
        }

        #endregion Public Functions

        //**************************************************************
        #region Public Events

        public event EventHandler<ImageActionEventArgs> ImageAction;

        protected virtual void OnImageAction(ImageActionEventArgs e)
        {
            EventHandler<ImageActionEventArgs> handler = ImageAction;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        public event EventHandler<EventArgs> HistogramDataAction;

        public virtual void OnHistogramDataAction(EventArgs e)
        {
            EventHandler<EventArgs> handler = HistogramDataAction;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        public event EventHandler<EventArgs> HistogramSettingsAction;

        protected virtual void OnHistogramSettingsAction(EventArgs e)
        {
            EventHandler<EventArgs> handler = HistogramSettingsAction;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        public event EventHandler<EventArgs> ImageDummyDataAction;

        protected virtual void OnImageDummyDataAction(EventArgs e)
        {
            EventHandler<EventArgs> handler = ImageDummyDataAction;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        public event EventHandler<EventArgs> MergedImageSettingsAction;

        protected virtual void OnMergedImageSettingsAction(EventArgs e)
        {
            EventHandler<EventArgs> handler = MergedImageSettingsAction;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        #endregion Public Events

        //**************************************************************
        #region Private Variables

        private static volatile ControlPanelViewModel _instance;
        private static object syncRoot = new Object();
        private FliProCameraManager fliProCameraManager = FliProCameraManager.Instance;
        private List<TabAction> TabList = new List<TabAction>();
        private GrabPanelViewModel GPVM = new GrabPanelViewModel();

        #endregion Private Variables

        //**************************************************************
        #region Private Functions

        #endregion Private Functions

    }
}
