﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using FliUtils.Jobs;
using System.Windows;
using System.IO;
using System.Diagnostics;
using System.Windows.Interop;
using System.Windows.Forms;
using System.Xml.Serialization;

namespace FliProCameraLib
{
    /// <summary>
    /// This file contains properties and methods that are only settable through the Factory dll
    /// for the FliProCamera class
    /// 2017/05/02
    /// </summary>
    public partial class FliProCamera : FliPro
    {
        //**************************************************************
        #region Constants and Enums

        public const int MAGIC_NUMBER = 0x4D657461;
        public const short NVSTORAGE_VERSION_1_HEADER_LENGTH = 32;
        public const short NVSTORAGE_VERSION_2_HEADER_LENGTH = 32;
        public const short NVSTORAGE_VERSION_3_HEADER_LENGTH = 32;
        public const short NVSTORAGE_VERSION_1 = 1;
        public const short NVSTORAGE_VERSION_2 = 2;
        public const short NVSTORAGE_VERSION_3 = 3;
        public const short NVSTORAGE_COUNTS_CLASS_LENGTH = GAIN_INDEX_SIZE + COUNTS_PER_ELECTRON_SIZE;
        public const short NVSTORAGE_MERGE_SLOPE_OFFSET_LENGTH = (2 * BLACK_LEVEL_SIZE) + (2 * GAIN_INDEX_SIZE) + GAIN_RATIO_SIZE +
            (3 * TEMPERATURE_SIZE) + (3 * SLOPE_OFFSET_SIZE);
        public const short NVTEMPERATURE_DIVISOR = 100;
        public const short NVGAIN_RATIO_DIVISOR = 1000;
        public const short NVCOUNTS_PER_ELECTRON_DIVISOR = 10000;
        public const short NVSTORAGE_MAX_CAPACITY = 1024;             // 1K bytes

        #endregion Constants and Enums

        //**************************************************************
        #region Offset Constants

        private const int MAGIC_NUMBER_SIZE = 4;                             // 4 Bytes					x4D657461	
        private const int NVSTORAGE_LENGTH_SIZE = 2;                         // 2 Bytes
        private const int NVSTORAGE_VERSION_SIZE = 2;                        // 2 Bytes
        private const int ADUS_PER_ELECTRON_SIZE = 4;                        // 4 Bytes
        private const int SLOPE_OFFSET_SIZE = 2;                             // 2 Bytes
        private const int COUNTS_PER_ELECTRON_SIZE = 4;                      // 4 Bytes
        private const int GAIN_RATIO_SIZE = 4;                               // 4 Bytes
        private const int TEMPERATURE_SIZE = 2;                              // 4 Bytes
        private const int BLACK_LEVEL_SIZE = 2;                              // 2 Bytes
        private const int GAIN_INDEX_SIZE = 2;                               // 2 Bytes
        private const int NUMBER_SIZE = 2;                                   // 2 Bytes

        #endregion Offset Constants

        //**************************************************************
        #region Protected Properties

        #endregion Protected Properties

        //**************************************************************
        #region Public Properties

        public long MagicNumber { get; set; }                           // 4	Bytes			0		x4D657461	
        public int NVLength { get; set; }                               // 2    Bytes           4
        public int NVVersion { get; set; }                              // 2    Bytes           6
        public double NVADUsPerElectron { get; set; }                   // 4    Bytes           8
        public double NVFactoryADUsPerElectron { get; set; }            // 4    Bytes          12
        public double NVGainRatio { get; set; }                         // 4    Bytes          16
        public double NVFactoryGainRatio { get; set; }                  // 4    Bytes          20
        public int NVNumberOfElectronsPerCountHighGains { get; set; }   // 2    Bytes          24
        public int NVNumberOfElectronsPerCountLowGains { get; set; }    // 2    Bytes          26
        public int NVNumberOfMergeLineOffsets { get; set; }             // 2    Bytes          28
        public int NVNumberOfFactoryMergeLineOffsets { get; set; }      // 2    Bytes          30

        #endregion Public Properties

        //**************************************************************
        #region Public Classes

        public class ElectronsPerCountClass
        {
            public ElectronsPerCountClass() { }
            public ElectronsPerCountClass(int index, double electronsPerCount)
            {
                Index = index;
                ElectronsPerCount = electronsPerCount;
            }

            public int Index { get; set; }

            [XmlIgnore]
            public double ElectronsPerCount { get; set; }

            [XmlElement("ElectronsPerCount")]
            public string ElectronsPerCountString
            {
                get { return ElectronsPerCount.ToString("0.0000"); }
                set { ElectronsPerCount = double.Parse(value); }
            }
        }

        public class MergeLineOffsetClass
        {
            public MergeLineOffsetClass() { }
            public MergeLineOffsetClass(uint blackLevelHigh, uint blackLevelLow, int highGainIndex, int lowGainIndex, double gainRatio, double temperature1, int lineOffset1,
                double temperature2, int lineOffset2, double temperature3, int lineOffset3)
            {
                BlackLevelHigh = blackLevelHigh;
                BlackLevelLow = blackLevelLow;
                HighGainIndex = highGainIndex;
                LowGainIndex = lowGainIndex;
                GainRatio = gainRatio;
                Temperature1 = temperature1;
                LineOffset1 = lineOffset1;
                Temperature2 = temperature2;
                LineOffset2 = lineOffset2;
                Temperature3 = temperature3;
                LineOffset3 = lineOffset3;
            }

            public uint BlackLevelHigh { get; set; }
            public uint BlackLevelLow { get; set; }
            public int HighGainIndex { get; set; }
            public int LowGainIndex { get; set; }

            [XmlIgnore]
            public double GainRatio { get; set; }

            [XmlElement("GainRatio")]
            public string GainRatioString
            {
                get { return GainRatio.ToString("0.000"); }
                set { GainRatio = double.Parse(value); }
            }

            [XmlIgnore]
            public double Temperature1 { get; set; }

            [XmlElement("Temperature1")]
            public string Temperature1String
            {
                get { return Temperature1.ToString("0.00"); }
                set { Temperature1 = double.Parse(value); }
            }

            public int LineOffset1 { get; set; }

            [XmlIgnore]
            public double Temperature2 { get; set; }

            [XmlElement("Temperature2")]
            public string Temperature2String
            {
                get { return Temperature2.ToString("0.00"); }
                set { Temperature2 = double.Parse(value); }
            }

            public int LineOffset2 { get; set; }

            [XmlIgnore]
            public double Temperature3 { get; set; }

            [XmlElement("Temperature3")]
            public string Temperature3String
            {
                get { return Temperature3.ToString("0.000"); }
                set { Temperature3 = double.Parse(value); }
            }

            public int LineOffset3 { get; set; }
        }

        public class MergeSettingsClass
        {
            public long MagicNumber { get; set; }                           // 4	Bytes			0		x4D657461	
            public int NVLength { get; set; }                               // 2    Bytes           4
            public int NVVersion { get; set; }                              // 2    Bytes           6
            public double NVADUsPerElectron { get; set; }                   // 4    Bytes           8
            public double NVFactoryADUsPerElectron { get; set; }            // 4    Bytes          12
            [XmlIgnore]
            public double NVGainRatio { get; set; }                         // 4    Bytes          16

            [XmlElement("NVGainRatio")]
            public string NVGainRatioString
            {
                get { return NVGainRatio.ToString("0.000"); }
                set { NVGainRatio = double.Parse(value); }
            }

            [XmlIgnore]
            public double NVFactoryGainRatio { get; set; }                  // 4    Bytes          20

            [XmlElement("NVFactoryGainRatio")]
            public string NVFactoryGainRatioString
            {
                get { return NVFactoryGainRatio.ToString("0.000"); }
                set { NVFactoryGainRatio = double.Parse(value); }
            }

            public int NVNumberOfElectronsPerCountHighGains { get; set; }   // 2    Bytes          24
            public int NVNumberOfElectronsPerCountLowGains { get; set; }    // 2    Bytes          26
            public int NVNumberOfMergeLineOffsets { get; set; }             // 2    Bytes          28
            public int NVNumberOfFactoryMergeLineOffsets { get; set; }      // 2    Bytes          30
            public MergeLineOffsetClass[] MergeLineOffsetList { get; set; }
            public MergeLineOffsetClass[] FactoryMergeLineOffsetList { get; set; }
            public ElectronsPerCountClass[] FactoryElectronsPerCountHighGainList { get; set; }
            public ElectronsPerCountClass[] FactoryElectronsPerCountLowGainList { get; set; }
        }

        #endregion Public Classes

        //**************************************************************
        #region Observable Properties
        #endregion Observable Properties

        //**************************************************************
        #region Public Collections

        /// <summary>
        /// This list is the one set by the customer to be used as thier defaults.
        /// </summary>
        public List<MergeLineOffsetClass> MergeLineOffsetList { get; set; }

        /// <summary>
        /// This list is the one set at the factory when calculating the best gain values for each temperature setting.
        /// </summary>
        public List<MergeLineOffsetClass> FactoryMergeLineOffsetList { get; set; }

        /// <summary>
        /// These are the values for each gain level saved at the factor. These will be used when calculating the ADUs for merge.
        /// </summary>
        public List<ElectronsPerCountClass> FactoryElectronsPerCountHighGainList { get; set; }
        public List<ElectronsPerCountClass> FactoryElectronsPerCountLowGainList { get; set; }

        #endregion Public Collections

        //**************************************************************
        #region Public Methods

        public void ClearAllMergeSettings()
        {
            MergeLineOffsetList.Clear();
            NVNumberOfMergeLineOffsets = 0;

            byte[] data = GetByteArray();

            WriteNVStorage(0, data, (uint)data.Length);
        }

        public void DeleteMergeSettings(int highGainIndex, int lowGainIndex)
        {
            int index = -1;

            for (int i = 0; i < MergeLineOffsetList.Count; i++)
            {
                if ((MergeLineOffsetList[i].HighGainIndex == highGainIndex) && (MergeLineOffsetList[i].LowGainIndex == lowGainIndex))
                {
                    index = i;
                    break;
                }
            }
            if (index >= 0)
            {
                MergeLineOffsetList.RemoveAt(index);
                NVNumberOfMergeLineOffsets = MergeLineOffsetList.Count;

                byte[] data = GetByteArray();

                WriteNVStorage(0, data, (uint)data.Length);
            }
        }

        public int GetFactoryElectronsPerCountHighIndex(int highGainIndex)
        {
            int index = -1;

            for (int i = 0; i < FactoryElectronsPerCountHighGainList.Count; i++)
            {
                if (FactoryElectronsPerCountHighGainList[i].Index == highGainIndex)
                {
                    index = i;
                    break;
                }
            }
            return index;
        }

        public int GetFactoryElectronsPerCountLowIndex(int lowGainIndex)
        {
            int index = -1;

            for (int i = 0; i < FactoryElectronsPerCountLowGainList.Count; i++)
            {
                if (FactoryElectronsPerCountLowGainList[i].Index == lowGainIndex)
                {
                    index = i;
                    break;
                }
            }
            return index;
        }

        public int GetFactoryMergeLineOffsetIndex(int highGainIndex, int lowGainIndex)
        {
            int index = -1;

            for (int i = 0; i < FactoryMergeLineOffsetList.Count; i++)
            {
                if ((FactoryMergeLineOffsetList[i].HighGainIndex == highGainIndex) && (FactoryMergeLineOffsetList[i].LowGainIndex == lowGainIndex))
                {
                    index = i;
                    break;
                }
            }
            if (index >= 0)
            {
                ADUsPerElectron = NVFactoryADUsPerElectron;
                GainRatio = FactoryMergeLineOffsetList[index].GainRatio;
                Temperature1 = FactoryMergeLineOffsetList[index].Temperature1;
                Temperature2 = FactoryMergeLineOffsetList[index].Temperature2;
                Temperature3 = FactoryMergeLineOffsetList[index].Temperature3;
                LineOffset1 = FactoryMergeLineOffsetList[index].LineOffset1;
                LineOffset2 = FactoryMergeLineOffsetList[index].LineOffset2;
                LineOffset3 = FactoryMergeLineOffsetList[index].LineOffset3;

                int tempIndex = GetFactoryElectronsPerCountHighIndex(highGainIndex);
                if (tempIndex >= 0)
                {
                    ActiveElectronsPerCountHigh = FactoryElectronsPerCountHighGainList[tempIndex].ElectronsPerCount;
                }
                tempIndex = GetFactoryElectronsPerCountLowIndex(highGainIndex);
                if (tempIndex >= 0)
                {
                    ActiveElectronsPerCountLow = FactoryElectronsPerCountLowGainList[tempIndex].ElectronsPerCount;
                }
            }
            return index;
        }

        public int GetMergeLineOffsetIndex(int highGainIndex, int lowGainIndex)
        {
            int index = -1;

            for (int i = 0; i < MergeLineOffsetList.Count; i++)
            {
                if ((MergeLineOffsetList[i].HighGainIndex == highGainIndex) && (MergeLineOffsetList[i].LowGainIndex == lowGainIndex))
                {
                    index = i;
                    break;
                }
            }
            if (index >= 0)
            {
                ADUsPerElectron = NVADUsPerElectron;
                GainRatio = MergeLineOffsetList[index].GainRatio;
                Temperature1 = MergeLineOffsetList[index].Temperature1;
                Temperature2 = MergeLineOffsetList[index].Temperature2;
                Temperature3 = MergeLineOffsetList[index].Temperature3;
                LineOffset1 = MergeLineOffsetList[index].LineOffset1;
                LineOffset2 = MergeLineOffsetList[index].LineOffset2;
                LineOffset3 = MergeLineOffsetList[index].LineOffset3;

                int tempIndex = GetFactoryElectronsPerCountHighIndex(highGainIndex);
                if (tempIndex >= 0)
                {
                    ActiveElectronsPerCountHigh = FactoryElectronsPerCountHighGainList[tempIndex].ElectronsPerCount;
                }
                tempIndex = GetFactoryElectronsPerCountLowIndex(highGainIndex);
                if (tempIndex >= 0)
                {
                    ActiveElectronsPerCountLow = FactoryElectronsPerCountLowGainList[tempIndex].ElectronsPerCount;
                }
            }
            return index;
        }

        public MergeSettingsClass GetCurrentMergeSettingsClass()
        {
            MergeSettingsClass mergeSettingsClass = new MergeSettingsClass();

            mergeSettingsClass.MagicNumber = MagicNumber;
            mergeSettingsClass.NVLength = NVLength;
            mergeSettingsClass.NVVersion = NVVersion;
            mergeSettingsClass.NVADUsPerElectron = NVADUsPerElectron;  
            mergeSettingsClass.NVFactoryADUsPerElectron = NVFactoryADUsPerElectron; 
            mergeSettingsClass.NVGainRatio = NVGainRatio;       
            mergeSettingsClass.NVFactoryGainRatio = NVFactoryGainRatio; 
            mergeSettingsClass.NVNumberOfElectronsPerCountHighGains = NVNumberOfElectronsPerCountHighGains;
            mergeSettingsClass.NVNumberOfElectronsPerCountLowGains = NVNumberOfElectronsPerCountLowGains;
            mergeSettingsClass.NVNumberOfMergeLineOffsets = NVNumberOfMergeLineOffsets;
            mergeSettingsClass.NVNumberOfFactoryMergeLineOffsets = NVNumberOfFactoryMergeLineOffsets;

            mergeSettingsClass.MergeLineOffsetList = new MergeLineOffsetClass[NVNumberOfMergeLineOffsets];
            for (int i = 0; i < NVNumberOfMergeLineOffsets; i++)
            {
                mergeSettingsClass.MergeLineOffsetList[i] = new MergeLineOffsetClass(
                    MergeLineOffsetList[i].BlackLevelHigh,
                    MergeLineOffsetList[i].BlackLevelLow,
                    MergeLineOffsetList[i].HighGainIndex,
                    MergeLineOffsetList[i].LowGainIndex,
                    MergeLineOffsetList[i].GainRatio,
                    MergeLineOffsetList[i].Temperature1,
                    MergeLineOffsetList[i].LineOffset1,
                    MergeLineOffsetList[i].Temperature2,
                    MergeLineOffsetList[i].LineOffset2,
                    MergeLineOffsetList[i].Temperature3,
                    MergeLineOffsetList[i].LineOffset3);
            }

            mergeSettingsClass.FactoryMergeLineOffsetList = new MergeLineOffsetClass[NVNumberOfFactoryMergeLineOffsets];
            for (int i = 0; i < NVNumberOfFactoryMergeLineOffsets; i++)
            {
                mergeSettingsClass.FactoryMergeLineOffsetList[i] = new MergeLineOffsetClass(
                    FactoryMergeLineOffsetList[i].BlackLevelHigh,
                    FactoryMergeLineOffsetList[i].BlackLevelLow,
                    FactoryMergeLineOffsetList[i].HighGainIndex,
                    FactoryMergeLineOffsetList[i].LowGainIndex,
                    FactoryMergeLineOffsetList[i].GainRatio,
                    FactoryMergeLineOffsetList[i].Temperature1,
                    FactoryMergeLineOffsetList[i].LineOffset1,
                    FactoryMergeLineOffsetList[i].Temperature2,
                    FactoryMergeLineOffsetList[i].LineOffset2,
                    FactoryMergeLineOffsetList[i].Temperature3,
                    FactoryMergeLineOffsetList[i].LineOffset3);
            }
            mergeSettingsClass.FactoryElectronsPerCountHighGainList = new ElectronsPerCountClass[NVNumberOfElectronsPerCountHighGains];
            for (int i = 0; i < NVNumberOfElectronsPerCountHighGains; i++)
            {
                mergeSettingsClass.FactoryElectronsPerCountHighGainList[i] = new ElectronsPerCountClass(
                    FactoryElectronsPerCountHighGainList[i].Index,
                    FactoryElectronsPerCountHighGainList[i].ElectronsPerCount);
            }
            mergeSettingsClass.FactoryElectronsPerCountLowGainList = new ElectronsPerCountClass[NVNumberOfElectronsPerCountLowGains];
            for (int i = 0; i < NVNumberOfElectronsPerCountLowGains; i++)
            {
                mergeSettingsClass.FactoryElectronsPerCountLowGainList[i] = new ElectronsPerCountClass(
                    FactoryElectronsPerCountLowGainList[i].Index,
                    FactoryElectronsPerCountLowGainList[i].ElectronsPerCount);
            }

            return mergeSettingsClass;
    }

    /// <summary>
    /// This function is called from the merge tab to retrieve the last defaults the customer has saved.
    /// </summary>
    public void ReadDefaultMergeSettings()
        {
            byte[] twoBytes = new byte[2];
            byte[] fourBytes = new byte[4];
            int currentIndex = 0;
            byte[] data = new byte[NVSTORAGE_VERSION_1_HEADER_LENGTH];

            RestoringMergeSettings = true;

            try
            {
                ReadNVStorage(0, data, (uint)NVSTORAGE_VERSION_1_HEADER_LENGTH);

                Array.Copy(data, currentIndex, fourBytes, 0, MAGIC_NUMBER_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(fourBytes);
                }
                MagicNumber = BitConverter.ToInt32(fourBytes, 0);
                currentIndex += MAGIC_NUMBER_SIZE;

                if (MagicNumber == MAGIC_NUMBER)
                {
                    // Get the actual length
                    Array.Copy(data, currentIndex, twoBytes, 0, NVSTORAGE_LENGTH_SIZE);
                    if (BitConverter.IsLittleEndian)
                    {
                        Array.Reverse(twoBytes);
                    }
                    NVLength = BitConverter.ToInt16(twoBytes, 0);

                    if (NVLength <= NVSTORAGE_VERSION_1_HEADER_LENGTH)
                    {
                        ADUsPerElectron = 0;
                        GainRatio = 0;
                        Temperature1 = -10;
                        Temperature2 = 5;
                        Temperature3 = 20;
                        LineOffset1 = 0;
                        LineOffset2 = 0;
                        LineOffset3 = 0;
                        ActiveElectronsPerCountLow = 0;
                        ActiveElectronsPerCountHigh = 0;
                        MergeLineOffsetList = new List<MergeLineOffsetClass>();
                        FactoryMergeLineOffsetList = new List<MergeLineOffsetClass>();
                        FactoryElectronsPerCountHighGainList = new List<ElectronsPerCountClass>();
                        FactoryElectronsPerCountLowGainList = new List<ElectronsPerCountClass>();
                        return;
                    }
                }
                else
                {
                    ADUsPerElectron = 0;
                    GainRatio = 0;
                    Temperature1 = -10;
                    Temperature2 = 5;
                    Temperature3 = 20;
                    LineOffset1 = 0;
                    LineOffset2 = 0;
                    LineOffset3 = 0;
                    ActiveElectronsPerCountLow = 0;
                    ActiveElectronsPerCountHigh = 0;
                    MergeLineOffsetList = new List<MergeLineOffsetClass>();
                    FactoryMergeLineOffsetList = new List<MergeLineOffsetClass>();
                    FactoryElectronsPerCountHighGainList = new List<ElectronsPerCountClass>();
                    FactoryElectronsPerCountLowGainList = new List<ElectronsPerCountClass>();
                    return;
                }

                data = new byte[NVLength];
                ReadNVStorage(0, data, (uint)NVLength);
                ParseNVStorage(data);

                ADUsPerElectron = NVADUsPerElectron;
                GainRatio = NVGainRatio;

                // For now we are only coding for the 4040
                if ((FactoryMergeLineOffsetList != null) && (FactoryMergeLineOffsetList.Count > 0))
                {
                    HighGainIndex = MergeLineOffsetList[0].HighGainIndex;
                    LowGainIndex = MergeLineOffsetList[0].LowGainIndex;
                    Temperature1 = MergeLineOffsetList[0].Temperature1;
                    Temperature2 = MergeLineOffsetList[0].Temperature2;
                    Temperature3 = MergeLineOffsetList[0].Temperature3;
                    LineOffset1 = MergeLineOffsetList[0].LineOffset1;
                    LineOffset2 = MergeLineOffsetList[0].LineOffset2;
                    LineOffset3 = MergeLineOffsetList[0].LineOffset3;
                    GainRatio = MergeLineOffsetList[0].GainRatio;
                }
                if ((FactoryElectronsPerCountLowGainList != null) && (FactoryElectronsPerCountLowGainList.Count > 0))
                {
                    ActiveElectronsPerCountLow = FactoryElectronsPerCountLowGainList[0].ElectronsPerCount;
                }
                if ((FactoryElectronsPerCountLowGainList != null) && (FactoryElectronsPerCountLowGainList.Count > 0))
                {
                    ActiveElectronsPerCountHigh = FactoryElectronsPerCountHighGainList[0].ElectronsPerCount;
                }
            }
            catch
            {
                EraseNVStorage();
                ADUsPerElectron = 0;
                GainRatio = 0;
                Temperature1 = -10;
                Temperature2 = 5;
                Temperature3 = 20;
                LineOffset1 = 0;
                LineOffset2 = 0;
                LineOffset3 = 0;
                ActiveElectronsPerCountLow = 0;
                ActiveElectronsPerCountHigh = 0;
                MergeLineOffsetList = new List<MergeLineOffsetClass>();
                FactoryMergeLineOffsetList = new List<MergeLineOffsetClass>();
                FactoryElectronsPerCountHighGainList = new List<ElectronsPerCountClass>();
                FactoryElectronsPerCountLowGainList = new List<ElectronsPerCountClass>();
            }

            RestoringMergeSettings = false;
        }

        /// <summary>
        /// This function is called from the merge tab to retrieve the factory settings.
        /// If the user wants to save these as his defaults he will do this as a seperate step.
        /// </summary>
        public void RestoreMergeSettingsFromFactorDefaults()
        {
            byte[] twoBytes = new byte[2];
            byte[] fourBytes = new byte[4];
            int currentIndex = 0;
            byte[] data = new byte[NVSTORAGE_VERSION_1_HEADER_LENGTH];

            RestoringMergeSettings = true;

            try
            {
                ReadNVStorage(0, data, (uint)NVSTORAGE_VERSION_1_HEADER_LENGTH);

                Array.Copy(data, currentIndex, fourBytes, 0, MAGIC_NUMBER_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(fourBytes);
                }
                MagicNumber = BitConverter.ToInt32(fourBytes, 0);
                currentIndex += MAGIC_NUMBER_SIZE;

                if (MagicNumber == MAGIC_NUMBER)
                {
                    // Get the actual length
                    Array.Copy(data, currentIndex, twoBytes, 0, NVSTORAGE_LENGTH_SIZE);
                    if (BitConverter.IsLittleEndian)
                    {
                        Array.Reverse(twoBytes);
                    }
                    NVLength = BitConverter.ToInt16(twoBytes, 0);

                    if (NVLength < NVSTORAGE_VERSION_1_HEADER_LENGTH)
                    {
                        ADUsPerElectron = 0;
                        GainRatio = 0;
                        Temperature1 = -10;
                        Temperature2 = 5;
                        Temperature3 = 20;
                        LineOffset1 = 0;
                        LineOffset2 = 0;
                        LineOffset3 = 0;
                        ActiveElectronsPerCountLow = 0;
                        ActiveElectronsPerCountHigh = 0;
                        MergeLineOffsetList = new List<MergeLineOffsetClass>();
                        FactoryMergeLineOffsetList = new List<MergeLineOffsetClass>();
                        FactoryElectronsPerCountHighGainList = new List<ElectronsPerCountClass>();
                        FactoryElectronsPerCountLowGainList = new List<ElectronsPerCountClass>();
                        return;
                    }
                }
                else
                {
                    ADUsPerElectron = 0;
                    GainRatio = 0;
                    Temperature1 = -10;
                    Temperature2 = 5;
                    Temperature3 = 20;
                    LineOffset1 = 0;
                    LineOffset2 = 0;
                    LineOffset3 = 0;
                    ActiveElectronsPerCountLow = 0;
                    ActiveElectronsPerCountHigh = 0;
                    MergeLineOffsetList = new List<MergeLineOffsetClass>();
                    FactoryMergeLineOffsetList = new List<MergeLineOffsetClass>();
                    FactoryElectronsPerCountHighGainList = new List<ElectronsPerCountClass>();
                    FactoryElectronsPerCountLowGainList = new List<ElectronsPerCountClass>();
                    return;
                }

                data = new byte[NVLength];
                ReadNVStorage(0, data, (uint)NVLength);
                ParseNVStorage(data);

                ADUsPerElectron = NVFactoryADUsPerElectron;
                GainRatio = NVFactoryGainRatio;

                // For now we are only coding for the 4040
                if ((FactoryMergeLineOffsetList != null) && (FactoryMergeLineOffsetList.Count > 0))
                {
                    BlackLevelHigh = FactoryMergeLineOffsetList[0].BlackLevelHigh;
                    BlackLevelLow = FactoryMergeLineOffsetList[0].BlackLevelLow;
                    HighGainIndex = FactoryMergeLineOffsetList[0].HighGainIndex;
                    LowGainIndex = FactoryMergeLineOffsetList[0].LowGainIndex;
                    Temperature1 = FactoryMergeLineOffsetList[0].Temperature1;
                    Temperature2 = FactoryMergeLineOffsetList[0].Temperature2;
                    Temperature3 = FactoryMergeLineOffsetList[0].Temperature3;
                    LineOffset1 = FactoryMergeLineOffsetList[0].LineOffset1;
                    LineOffset2 = FactoryMergeLineOffsetList[0].LineOffset2;
                    LineOffset3 = FactoryMergeLineOffsetList[0].LineOffset3;
                    GainRatio = FactoryMergeLineOffsetList[0].GainRatio;
                }
                if ((FactoryElectronsPerCountLowGainList != null) && (FactoryElectronsPerCountLowGainList.Count > 0))
                {
                    ActiveElectronsPerCountLow = FactoryElectronsPerCountLowGainList[0].ElectronsPerCount;
                }
                if ((FactoryElectronsPerCountLowGainList != null) && (FactoryElectronsPerCountLowGainList.Count > 0))
                {
                    ActiveElectronsPerCountHigh = FactoryElectronsPerCountHighGainList[0].ElectronsPerCount;
                }
                OnMergeValuesChangeAction(new EventArgs());
            }
            catch
            {
                EraseNVStorage();
                ADUsPerElectron = 0;
                GainRatio = 0;
                Temperature1 = -10;
                Temperature2 = 5;
                Temperature3 = 20;
                LineOffset1 = 0;
                LineOffset2 = 0;
                LineOffset3 = 0;
                ActiveElectronsPerCountLow = 0;
                ActiveElectronsPerCountHigh = 0;
                MergeLineOffsetList = new List<MergeLineOffsetClass>();
                FactoryMergeLineOffsetList = new List<MergeLineOffsetClass>();
                FactoryElectronsPerCountHighGainList = new List<ElectronsPerCountClass>();
                FactoryElectronsPerCountLowGainList = new List<ElectronsPerCountClass>();
            }

            RestoringMergeSettings = false;
        }

        public bool SaveMergeSettingsFromFile(MergeSettingsClass mergeSettingsClass)
        {
            bool isValid = ((mergeSettingsClass.MergeLineOffsetList.Length == mergeSettingsClass.NVNumberOfMergeLineOffsets) &&
                (mergeSettingsClass.FactoryMergeLineOffsetList.Length == mergeSettingsClass.NVNumberOfFactoryMergeLineOffsets) &&
                (mergeSettingsClass.FactoryElectronsPerCountHighGainList.Length == mergeSettingsClass.NVNumberOfElectronsPerCountHighGains) &&
                (mergeSettingsClass.FactoryElectronsPerCountLowGainList.Length == mergeSettingsClass.NVNumberOfElectronsPerCountLowGains));

            if (isValid)
            {
                MagicNumber = mergeSettingsClass.MagicNumber;
                NVLength = mergeSettingsClass.NVLength;
                NVVersion = mergeSettingsClass.NVVersion;
                NVADUsPerElectron = mergeSettingsClass.NVADUsPerElectron;
                NVFactoryADUsPerElectron = mergeSettingsClass.NVFactoryADUsPerElectron;
                NVGainRatio = mergeSettingsClass.NVGainRatio;
                NVFactoryGainRatio = mergeSettingsClass.NVFactoryGainRatio;
                NVNumberOfElectronsPerCountHighGains = mergeSettingsClass.NVNumberOfElectronsPerCountHighGains;
                NVNumberOfElectronsPerCountLowGains = mergeSettingsClass.NVNumberOfElectronsPerCountLowGains;
                NVNumberOfMergeLineOffsets = mergeSettingsClass.NVNumberOfMergeLineOffsets;
                NVNumberOfFactoryMergeLineOffsets = mergeSettingsClass.NVNumberOfFactoryMergeLineOffsets;

                MergeLineOffsetList = new List<MergeLineOffsetClass>();
                for (int i = 0; i < mergeSettingsClass.MergeLineOffsetList.Length; i++)
                {
                    MergeLineOffsetClass mergeLineOffsetClass = new MergeLineOffsetClass(
                        mergeSettingsClass.MergeLineOffsetList[i].BlackLevelHigh,
                        mergeSettingsClass.MergeLineOffsetList[i].BlackLevelLow,
                        mergeSettingsClass.MergeLineOffsetList[i].HighGainIndex,
                        mergeSettingsClass.MergeLineOffsetList[i].LowGainIndex,
                        mergeSettingsClass.MergeLineOffsetList[i].GainRatio,
                        mergeSettingsClass.MergeLineOffsetList[i].Temperature1,
                        mergeSettingsClass.MergeLineOffsetList[i].LineOffset1,
                        mergeSettingsClass.MergeLineOffsetList[i].Temperature2,
                        mergeSettingsClass.MergeLineOffsetList[i].LineOffset2,
                        mergeSettingsClass.MergeLineOffsetList[i].Temperature3,
                        mergeSettingsClass.MergeLineOffsetList[i].LineOffset3);
                    MergeLineOffsetList.Add(mergeLineOffsetClass);
                }

                FactoryMergeLineOffsetList = new List<MergeLineOffsetClass>();
                for (int i = 0; i < mergeSettingsClass.FactoryMergeLineOffsetList.Length; i++)
                {
                    MergeLineOffsetClass mergeLineOffsetClass = new MergeLineOffsetClass(
                        mergeSettingsClass.FactoryMergeLineOffsetList[i].BlackLevelHigh,
                        mergeSettingsClass.FactoryMergeLineOffsetList[i].BlackLevelLow,
                        mergeSettingsClass.FactoryMergeLineOffsetList[i].HighGainIndex,
                        mergeSettingsClass.FactoryMergeLineOffsetList[i].LowGainIndex,
                        mergeSettingsClass.FactoryMergeLineOffsetList[i].GainRatio,
                        mergeSettingsClass.FactoryMergeLineOffsetList[i].Temperature1,
                        mergeSettingsClass.FactoryMergeLineOffsetList[i].LineOffset1,
                        mergeSettingsClass.FactoryMergeLineOffsetList[i].Temperature2,
                        mergeSettingsClass.FactoryMergeLineOffsetList[i].LineOffset2,
                        mergeSettingsClass.FactoryMergeLineOffsetList[i].Temperature3,
                        mergeSettingsClass.FactoryMergeLineOffsetList[i].LineOffset3);
                    FactoryMergeLineOffsetList.Add(mergeLineOffsetClass);
                }

                FactoryElectronsPerCountHighGainList = new List<ElectronsPerCountClass>();
                for (int i = 0; i < NVNumberOfElectronsPerCountHighGains; i++)
                {
                    ElectronsPerCountClass electronsPerCountClass = new ElectronsPerCountClass(
                        mergeSettingsClass.FactoryElectronsPerCountHighGainList[i].Index,
                        mergeSettingsClass.FactoryElectronsPerCountHighGainList[i].ElectronsPerCount);
                    FactoryElectronsPerCountHighGainList.Add(electronsPerCountClass);
                }

                FactoryElectronsPerCountLowGainList = new List<ElectronsPerCountClass>();
                for (int i = 0; i < mergeSettingsClass.FactoryElectronsPerCountLowGainList.Length; i++)
                {
                    ElectronsPerCountClass electronsPerCountClass = new ElectronsPerCountClass(
                        mergeSettingsClass.FactoryElectronsPerCountLowGainList[i].Index,
                        mergeSettingsClass.FactoryElectronsPerCountLowGainList[i].ElectronsPerCount);
                    FactoryElectronsPerCountLowGainList.Add(electronsPerCountClass);
                }

                byte[] data = GetByteArray();

                WriteNVStorage(0, data, (uint)data.Length);

            }

            return isValid;
        }

        /// <summary>
        /// This function is called from the test routines so we want to initialize the customer defaults
        /// with the factory settings. If this is not correct we need to comment out the MergeLineOffsetList block.
        /// </summary>
        public void SaveMergeSettingsAsFactorDefaults()
        {
            // Set the values to be stored to the target values
            NVADUsPerElectron = ADUsPerElectron;
            NVFactoryADUsPerElectron = ADUsPerElectron;
            GainRatio = Math.Round(GainRatio, 4);
            NVGainRatio = GainRatio;
            NVFactoryGainRatio = GainRatio;

            // Add the current settings to the list
            UpdateMergeLineOffsetList(MergeLineOffsetList, BlackLevelHigh, BlackLevelLow, HighGainIndex, LowGainIndex, GainRatio, Temperature1, LineOffset1,
                Temperature2, LineOffset2, Temperature3, LineOffset3);
            NVNumberOfMergeLineOffsets = MergeLineOffsetList.Count;

            UpdateMergeLineOffsetList(FactoryMergeLineOffsetList, BlackLevelHigh, BlackLevelLow, HighGainIndex, LowGainIndex, GainRatio, Temperature1, LineOffset1,
                Temperature2, LineOffset2, Temperature3, LineOffset3);
            NVNumberOfFactoryMergeLineOffsets = FactoryMergeLineOffsetList.Count;

            UpdateElectronsPerCountList(FactoryElectronsPerCountHighGainList, HighGainIndex, ActiveElectronsPerCountHigh);
            NVNumberOfElectronsPerCountHighGains = FactoryElectronsPerCountHighGainList.Count;

            UpdateElectronsPerCountList(FactoryElectronsPerCountLowGainList, LowGainIndex, ActiveElectronsPerCountLow);
            NVNumberOfElectronsPerCountLowGains = FactoryElectronsPerCountLowGainList.Count;

            byte[] data = GetByteArray();

            WriteNVStorage(0, data, (uint)data.Length);
        }

        /// <summary>
        /// This is called from the Merge tab when the customer wants to save their values as the defaults.
        /// </summary>
        public void StoreNVDataToCamera()
        {
            int dataLength = NVSTORAGE_VERSION_3_HEADER_LENGTH + (FactoryElectronsPerCountHighGainList.Count * NVSTORAGE_COUNTS_CLASS_LENGTH) +
                (FactoryElectronsPerCountLowGainList.Count * NVSTORAGE_COUNTS_CLASS_LENGTH) +
                (MergeLineOffsetList.Count * NVSTORAGE_MERGE_SLOPE_OFFSET_LENGTH) +
                (FactoryMergeLineOffsetList.Count * NVSTORAGE_MERGE_SLOPE_OFFSET_LENGTH);
            int index = -1;

            for (int i = 0; i < MergeLineOffsetList.Count; i++)
            {
                if ((MergeLineOffsetList[i].HighGainIndex == HighGainIndex) && (MergeLineOffsetList[i].LowGainIndex == LowGainIndex))
                {
                    index = i;
                    break;
                }
            }
            if ((index >= 0) && (dataLength + NVSTORAGE_MERGE_SLOPE_OFFSET_LENGTH >= NVSTORAGE_MAX_CAPACITY))
            {
                System.Windows.MessageBox.Show("There is not enough room to store this entry.\nPlease Clear All saved values or\n" + 
                    "Delete the saved settings for a different High/Low Gain pair.");
                return;
            }

            // Set the values to be stored to the target values
            NVADUsPerElectron = ADUsPerElectron;
            GainRatio = Math.Round(GainRatio, 4);
            NVGainRatio = GainRatio;

            // Add the current settings to the list
            UpdateMergeLineOffsetList(MergeLineOffsetList, BlackLevelHigh, BlackLevelLow, HighGainIndex, LowGainIndex, GainRatio, Temperature1, LineOffset1,
                Temperature2, LineOffset2, Temperature3, LineOffset3);
            NVNumberOfMergeLineOffsets = MergeLineOffsetList.Count;

            // The FactoryMergeLineOffsetList, FactoryElectronsPerCountHighGainList, and FactoryElectronsPerCountLowGainList should already
            // be populated. If they are not we are in trouble.

            byte[] data = GetByteArray();

            WriteNVStorage(0, data, (uint)data.Length);
        }

        #endregion Public Methods

        //**************************************************************
        #region Protected Methods

        public void EraseNVStorage()
        {
            byte[] data = new byte[NVSTORAGE_VERSION_1_HEADER_LENGTH];

            for (int i = 0; i < data.Length; i++)
            {
                data[i] = 0xFF;
            }

            WriteNVStorage(0, data, (uint)data.Length);
        }

        public byte[] GetByteArray()
        {
            int currentIndex = 0;
            int dataLength = NVSTORAGE_VERSION_3_HEADER_LENGTH + (FactoryElectronsPerCountHighGainList.Count * NVSTORAGE_COUNTS_CLASS_LENGTH) +
                (FactoryElectronsPerCountLowGainList.Count * NVSTORAGE_COUNTS_CLASS_LENGTH) +
                (MergeLineOffsetList.Count * NVSTORAGE_MERGE_SLOPE_OFFSET_LENGTH) +
                (FactoryMergeLineOffsetList.Count * NVSTORAGE_MERGE_SLOPE_OFFSET_LENGTH);
            byte[] data = new byte[dataLength];

            StoreValue(MAGIC_NUMBER, data, ref currentIndex, MAGIC_NUMBER_SIZE);
            StoreValue(dataLength, data, ref currentIndex, NVSTORAGE_LENGTH_SIZE);
            StoreValue(NVSTORAGE_VERSION_3, data, ref currentIndex, NVSTORAGE_VERSION_SIZE);
            StoreValue(NVADUsPerElectron * NVGAIN_RATIO_DIVISOR, data, ref currentIndex, ADUS_PER_ELECTRON_SIZE);
            StoreValue(NVFactoryADUsPerElectron * NVGAIN_RATIO_DIVISOR, data, ref currentIndex, ADUS_PER_ELECTRON_SIZE);
            StoreValue(NVGainRatio * NVGAIN_RATIO_DIVISOR, data, ref currentIndex, GAIN_RATIO_SIZE);
            StoreValue(NVFactoryGainRatio * NVGAIN_RATIO_DIVISOR, data, ref currentIndex, GAIN_RATIO_SIZE);
            StoreValue(NVNumberOfElectronsPerCountHighGains, data, ref currentIndex, NUMBER_SIZE);
            StoreValue(NVNumberOfElectronsPerCountLowGains, data, ref currentIndex, NUMBER_SIZE);
            StoreValue(NVNumberOfMergeLineOffsets, data, ref currentIndex, NUMBER_SIZE);
            StoreValue(NVNumberOfFactoryMergeLineOffsets, data, ref currentIndex, NUMBER_SIZE);

            for (int i = 0; i < FactoryElectronsPerCountHighGainList.Count; i++)
            {
                StoreValue(FactoryElectronsPerCountHighGainList[i].Index, data, ref currentIndex, GAIN_INDEX_SIZE);
                StoreValue(FactoryElectronsPerCountHighGainList[i].ElectronsPerCount * NVCOUNTS_PER_ELECTRON_DIVISOR, data, ref currentIndex, COUNTS_PER_ELECTRON_SIZE);
            }
            for (int i = 0; i < FactoryElectronsPerCountLowGainList.Count; i++)
            {
                StoreValue(FactoryElectronsPerCountLowGainList[i].Index, data, ref currentIndex, GAIN_INDEX_SIZE);
                StoreValue(FactoryElectronsPerCountLowGainList[i].ElectronsPerCount * NVCOUNTS_PER_ELECTRON_DIVISOR, data, ref currentIndex, COUNTS_PER_ELECTRON_SIZE);
            }

            for (int i = 0; i < MergeLineOffsetList.Count; i++)
            {
                StoreValue(MergeLineOffsetList[i].BlackLevelHigh, data, ref currentIndex, BLACK_LEVEL_SIZE);
                StoreValue(MergeLineOffsetList[i].BlackLevelLow, data, ref currentIndex, BLACK_LEVEL_SIZE);
                StoreValue(MergeLineOffsetList[i].HighGainIndex, data, ref currentIndex, GAIN_INDEX_SIZE);
                StoreValue(MergeLineOffsetList[i].LowGainIndex, data, ref currentIndex, GAIN_INDEX_SIZE);
                StoreValue(MergeLineOffsetList[i].GainRatio * NVGAIN_RATIO_DIVISOR, data, ref currentIndex, GAIN_RATIO_SIZE);
                StoreValue(MergeLineOffsetList[i].Temperature1 * NVTEMPERATURE_DIVISOR, data, ref currentIndex, TEMPERATURE_SIZE);
                StoreValue(MergeLineOffsetList[i].LineOffset1, data, ref currentIndex, SLOPE_OFFSET_SIZE);
                StoreValue(MergeLineOffsetList[i].Temperature2 * NVTEMPERATURE_DIVISOR, data, ref currentIndex, TEMPERATURE_SIZE);
                StoreValue(MergeLineOffsetList[i].LineOffset2, data, ref currentIndex, SLOPE_OFFSET_SIZE);
                StoreValue(MergeLineOffsetList[i].Temperature3 * NVTEMPERATURE_DIVISOR, data, ref currentIndex, TEMPERATURE_SIZE);
                StoreValue(MergeLineOffsetList[i].LineOffset3, data, ref currentIndex, SLOPE_OFFSET_SIZE);
            }

            for (int i = 0; i < FactoryMergeLineOffsetList.Count; i++)
            {
                StoreValue(FactoryMergeLineOffsetList[i].BlackLevelHigh, data, ref currentIndex, BLACK_LEVEL_SIZE);
                StoreValue(FactoryMergeLineOffsetList[i].BlackLevelLow, data, ref currentIndex, BLACK_LEVEL_SIZE);
                StoreValue(FactoryMergeLineOffsetList[i].HighGainIndex, data, ref currentIndex, GAIN_INDEX_SIZE);
                StoreValue(FactoryMergeLineOffsetList[i].LowGainIndex, data, ref currentIndex, GAIN_INDEX_SIZE);
                StoreValue(FactoryMergeLineOffsetList[i].GainRatio * NVGAIN_RATIO_DIVISOR, data, ref currentIndex, GAIN_RATIO_SIZE);
                StoreValue(FactoryMergeLineOffsetList[i].Temperature1 * NVTEMPERATURE_DIVISOR, data, ref currentIndex, TEMPERATURE_SIZE);
                StoreValue(FactoryMergeLineOffsetList[i].LineOffset1, data, ref currentIndex, SLOPE_OFFSET_SIZE);
                StoreValue(FactoryMergeLineOffsetList[i].Temperature2 * NVTEMPERATURE_DIVISOR, data, ref currentIndex, TEMPERATURE_SIZE);
                StoreValue(FactoryMergeLineOffsetList[i].LineOffset2, data, ref currentIndex, SLOPE_OFFSET_SIZE);
                StoreValue(FactoryMergeLineOffsetList[i].Temperature3 * NVTEMPERATURE_DIVISOR, data, ref currentIndex, TEMPERATURE_SIZE);
                StoreValue(FactoryMergeLineOffsetList[i].LineOffset3, data, ref currentIndex, SLOPE_OFFSET_SIZE);
            }

            return data;
        }

        private void ParseNVStorage(byte[] data)
        {
            byte[] twoBytes = new byte[2];
            byte[] fourBytes = new byte[4];
            int currentIndex = 0;

            Array.Copy(data, currentIndex, fourBytes, 0, MAGIC_NUMBER_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            MagicNumber = BitConverter.ToInt32(fourBytes, 0);
            currentIndex += MAGIC_NUMBER_SIZE;

            Array.Copy(data, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            NVLength = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += NVSTORAGE_LENGTH_SIZE;

            Array.Copy(data, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            NVVersion = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += NVSTORAGE_VERSION_SIZE;

            if (NVVersion == NVSTORAGE_VERSION_1)
            {
                ParseNVStorageVersion1(data, currentIndex);

                // We no longer want to use Version 1 so write it back out now as Version 3
                data = GetByteArray();
                WriteNVStorage(0, data, (uint)data.Length);

            }
            else if (NVVersion == NVSTORAGE_VERSION_2)
            {
                ParseNVStorageVersion2(data, currentIndex);

                // We no longer want to use Version 2 so write it back out now as Version 3
                data = GetByteArray();
                WriteNVStorage(0, data, (uint)data.Length);
            }
            else if (NVVersion == NVSTORAGE_VERSION_3)
            {
                ParseNVStorageVersion3(data, currentIndex);
            }
        }

        private void ParseNVStorageVersion1(byte[] data, int currentIndex)
        {
            byte[] twoBytes = new byte[2];
            byte[] fourBytes = new byte[4];

            Array.Copy(data, currentIndex, fourBytes, 0, ADUS_PER_ELECTRON_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            NVADUsPerElectron = (double)BitConverter.ToUInt32(fourBytes, 0) / NVGAIN_RATIO_DIVISOR;
            currentIndex += ADUS_PER_ELECTRON_SIZE;

            Array.Copy(data, currentIndex, fourBytes, 0, ADUS_PER_ELECTRON_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            NVFactoryADUsPerElectron = (double)BitConverter.ToUInt32(fourBytes, 0) / NVGAIN_RATIO_DIVISOR;
            currentIndex += ADUS_PER_ELECTRON_SIZE;

            Array.Copy(data, currentIndex, fourBytes, 0, GAIN_RATIO_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            NVGainRatio = (double)BitConverter.ToUInt32(fourBytes, 0) / NVGAIN_RATIO_DIVISOR;
            currentIndex += GAIN_RATIO_SIZE;

            Array.Copy(data, currentIndex, fourBytes, 0, GAIN_RATIO_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            NVFactoryGainRatio = (double)BitConverter.ToUInt32(fourBytes, 0) / NVGAIN_RATIO_DIVISOR;
            currentIndex += GAIN_RATIO_SIZE;

            Array.Copy(data, currentIndex, twoBytes, 0, NUMBER_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            NVNumberOfElectronsPerCountHighGains = BitConverter.ToInt16(twoBytes, 0);
            currentIndex += NUMBER_SIZE;

            Array.Copy(data, currentIndex, twoBytes, 0, NUMBER_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            NVNumberOfElectronsPerCountLowGains = BitConverter.ToInt16(twoBytes, 0);
            currentIndex += NUMBER_SIZE;

            Array.Copy(data, currentIndex, twoBytes, 0, NUMBER_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            NVNumberOfMergeLineOffsets = BitConverter.ToInt16(twoBytes, 0);
            currentIndex += NUMBER_SIZE;

            Array.Copy(data, currentIndex, twoBytes, 0, NUMBER_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            NVNumberOfFactoryMergeLineOffsets = BitConverter.ToInt16(twoBytes, 0);
            currentIndex += NUMBER_SIZE;

            FactoryElectronsPerCountHighGainList = new List<ElectronsPerCountClass>();
            for (int i = 0; i < NVNumberOfElectronsPerCountHighGains; i++)
            {
                ElectronsPerCountClass electronsPerCountClass = new ElectronsPerCountClass();

                Array.Copy(data, currentIndex, twoBytes, 0, GAIN_INDEX_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                electronsPerCountClass.Index = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += GAIN_INDEX_SIZE;

                Array.Copy(data, currentIndex, fourBytes, 0, COUNTS_PER_ELECTRON_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(fourBytes);
                }
                electronsPerCountClass.ElectronsPerCount = (double)BitConverter.ToUInt32(fourBytes, 0) / NVCOUNTS_PER_ELECTRON_DIVISOR;
                currentIndex += COUNTS_PER_ELECTRON_SIZE;

                FactoryElectronsPerCountHighGainList.Add(electronsPerCountClass);
            }

            FactoryElectronsPerCountLowGainList = new List<ElectronsPerCountClass>();
            for (int i = 0; i < NVNumberOfElectronsPerCountLowGains; i++)
            {
                ElectronsPerCountClass electronsPerCountClass = new ElectronsPerCountClass();

                Array.Copy(data, currentIndex, twoBytes, 0, GAIN_INDEX_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                electronsPerCountClass.Index = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += GAIN_INDEX_SIZE;

                Array.Copy(data, currentIndex, fourBytes, 0, COUNTS_PER_ELECTRON_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(fourBytes);
                }
                electronsPerCountClass.ElectronsPerCount = (double)BitConverter.ToUInt32(fourBytes, 0) / NVCOUNTS_PER_ELECTRON_DIVISOR;
                currentIndex += COUNTS_PER_ELECTRON_SIZE;

                FactoryElectronsPerCountLowGainList.Add(electronsPerCountClass);
            }

            MergeLineOffsetList = new List<MergeLineOffsetClass>();
            for (int i = 0; i < NVNumberOfMergeLineOffsets; i++)
            {
                MergeLineOffsetClass mergeLineOffsetClass = new MergeLineOffsetClass();

                mergeLineOffsetClass.BlackLevelHigh = BlackLevelHigh;
                mergeLineOffsetClass.BlackLevelLow = BlackLevelLow;

                Array.Copy(data, currentIndex, twoBytes, 0, GAIN_INDEX_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.HighGainIndex = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += GAIN_INDEX_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, GAIN_INDEX_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LowGainIndex = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += GAIN_INDEX_SIZE;

                mergeLineOffsetClass.GainRatio = NVGainRatio;

                Array.Copy(data, currentIndex, twoBytes, 0, TEMPERATURE_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.Temperature1 = (double)BitConverter.ToInt16(twoBytes, 0) / NVTEMPERATURE_DIVISOR;
                currentIndex += TEMPERATURE_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, SLOPE_OFFSET_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LineOffset1 = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += SLOPE_OFFSET_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, TEMPERATURE_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.Temperature2 = (double)BitConverter.ToInt16(twoBytes, 0) / NVTEMPERATURE_DIVISOR;
                currentIndex += TEMPERATURE_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, SLOPE_OFFSET_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LineOffset2 = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += SLOPE_OFFSET_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, TEMPERATURE_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.Temperature3 = (double)BitConverter.ToInt16(twoBytes, 0) / NVTEMPERATURE_DIVISOR;
                currentIndex += TEMPERATURE_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, SLOPE_OFFSET_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LineOffset3 = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += SLOPE_OFFSET_SIZE;

                MergeLineOffsetList.Add(mergeLineOffsetClass);
            }

            FactoryMergeLineOffsetList = new List<MergeLineOffsetClass>();
            for (int i = 0; i < NVNumberOfFactoryMergeLineOffsets; i++)
            {
                MergeLineOffsetClass mergeLineOffsetClass = new MergeLineOffsetClass();

                mergeLineOffsetClass.BlackLevelHigh = BlackLevelHigh;
                mergeLineOffsetClass.BlackLevelLow = BlackLevelLow;

                Array.Copy(data, currentIndex, twoBytes, 0, GAIN_INDEX_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.HighGainIndex = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += GAIN_INDEX_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, GAIN_INDEX_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LowGainIndex = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += GAIN_INDEX_SIZE;

                mergeLineOffsetClass.GainRatio = NVGainRatio;

                Array.Copy(data, currentIndex, twoBytes, 0, TEMPERATURE_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.Temperature1 = (double)BitConverter.ToInt16(twoBytes, 0) / NVTEMPERATURE_DIVISOR;
                currentIndex += TEMPERATURE_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, SLOPE_OFFSET_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LineOffset1 = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += SLOPE_OFFSET_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, TEMPERATURE_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.Temperature2 = (double)BitConverter.ToInt16(twoBytes, 0) / NVTEMPERATURE_DIVISOR;
                currentIndex += TEMPERATURE_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, SLOPE_OFFSET_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LineOffset2 = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += SLOPE_OFFSET_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, TEMPERATURE_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.Temperature3 = (double)BitConverter.ToInt16(twoBytes, 0) / NVTEMPERATURE_DIVISOR;
                currentIndex += TEMPERATURE_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, SLOPE_OFFSET_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LineOffset3 = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += SLOPE_OFFSET_SIZE;

                FactoryMergeLineOffsetList.Add(mergeLineOffsetClass);
            }
        }

        private void ParseNVStorageVersion2(byte[] data, int currentIndex)
        {
            byte[] twoBytes = new byte[2];
            byte[] fourBytes = new byte[4];

            Array.Copy(data, currentIndex, fourBytes, 0, ADUS_PER_ELECTRON_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            NVADUsPerElectron = (double)BitConverter.ToUInt32(fourBytes, 0) / NVGAIN_RATIO_DIVISOR;
            currentIndex += ADUS_PER_ELECTRON_SIZE;

            Array.Copy(data, currentIndex, fourBytes, 0, ADUS_PER_ELECTRON_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            NVFactoryADUsPerElectron = (double)BitConverter.ToUInt32(fourBytes, 0) / NVGAIN_RATIO_DIVISOR;
            currentIndex += ADUS_PER_ELECTRON_SIZE;

            Array.Copy(data, currentIndex, fourBytes, 0, GAIN_RATIO_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            NVGainRatio = (double)BitConverter.ToUInt32(fourBytes, 0) / NVGAIN_RATIO_DIVISOR;
            currentIndex += GAIN_RATIO_SIZE;

            Array.Copy(data, currentIndex, fourBytes, 0, GAIN_RATIO_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            NVFactoryGainRatio = (double)BitConverter.ToUInt32(fourBytes, 0) / NVGAIN_RATIO_DIVISOR;
            currentIndex += GAIN_RATIO_SIZE;

            Array.Copy(data, currentIndex, twoBytes, 0, NUMBER_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            NVNumberOfElectronsPerCountHighGains = BitConverter.ToInt16(twoBytes, 0);
            currentIndex += NUMBER_SIZE;

            Array.Copy(data, currentIndex, twoBytes, 0, NUMBER_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            NVNumberOfElectronsPerCountLowGains = BitConverter.ToInt16(twoBytes, 0);
            currentIndex += NUMBER_SIZE;

            Array.Copy(data, currentIndex, twoBytes, 0, NUMBER_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            NVNumberOfMergeLineOffsets = BitConverter.ToInt16(twoBytes, 0);
            currentIndex += NUMBER_SIZE;

            Array.Copy(data, currentIndex, twoBytes, 0, NUMBER_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            NVNumberOfFactoryMergeLineOffsets = BitConverter.ToInt16(twoBytes, 0);
            currentIndex += NUMBER_SIZE;

            FactoryElectronsPerCountHighGainList = new List<ElectronsPerCountClass>();
            for (int i = 0; i < NVNumberOfElectronsPerCountHighGains; i++)
            {
                ElectronsPerCountClass electronsPerCountClass = new ElectronsPerCountClass();

                Array.Copy(data, currentIndex, twoBytes, 0, GAIN_INDEX_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                electronsPerCountClass.Index = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += GAIN_INDEX_SIZE;

                Array.Copy(data, currentIndex, fourBytes, 0, COUNTS_PER_ELECTRON_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(fourBytes);
                }
                electronsPerCountClass.ElectronsPerCount = (double)BitConverter.ToUInt32(fourBytes, 0) / NVCOUNTS_PER_ELECTRON_DIVISOR;
                currentIndex += COUNTS_PER_ELECTRON_SIZE;

                FactoryElectronsPerCountHighGainList.Add(electronsPerCountClass);
            }

            FactoryElectronsPerCountLowGainList = new List<ElectronsPerCountClass>();
            for (int i = 0; i < NVNumberOfElectronsPerCountLowGains; i++)
            {
                ElectronsPerCountClass electronsPerCountClass = new ElectronsPerCountClass();

                Array.Copy(data, currentIndex, twoBytes, 0, GAIN_INDEX_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                electronsPerCountClass.Index = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += GAIN_INDEX_SIZE;

                Array.Copy(data, currentIndex, fourBytes, 0, COUNTS_PER_ELECTRON_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(fourBytes);
                }
                electronsPerCountClass.ElectronsPerCount = (double)BitConverter.ToUInt32(fourBytes, 0) / NVCOUNTS_PER_ELECTRON_DIVISOR;
                currentIndex += COUNTS_PER_ELECTRON_SIZE;

                FactoryElectronsPerCountLowGainList.Add(electronsPerCountClass);
            }

            MergeLineOffsetList = new List<MergeLineOffsetClass>();
            for (int i = 0; i < NVNumberOfMergeLineOffsets; i++)
            {
                MergeLineOffsetClass mergeLineOffsetClass = new MergeLineOffsetClass();

                mergeLineOffsetClass.BlackLevelHigh = BlackLevelHigh;
                mergeLineOffsetClass.BlackLevelLow = BlackLevelLow;

                Array.Copy(data, currentIndex, twoBytes, 0, GAIN_INDEX_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.HighGainIndex = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += GAIN_INDEX_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, GAIN_INDEX_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LowGainIndex = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += GAIN_INDEX_SIZE;

                Array.Copy(data, currentIndex, fourBytes, 0, GAIN_RATIO_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(fourBytes);
                }
                mergeLineOffsetClass.GainRatio = (double)BitConverter.ToUInt32(fourBytes, 0) / NVGAIN_RATIO_DIVISOR;
                currentIndex += GAIN_RATIO_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, TEMPERATURE_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.Temperature1 = (double)BitConverter.ToInt16(twoBytes, 0) / NVTEMPERATURE_DIVISOR;
                currentIndex += TEMPERATURE_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, SLOPE_OFFSET_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LineOffset1 = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += SLOPE_OFFSET_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, TEMPERATURE_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.Temperature2 = (double)BitConverter.ToInt16(twoBytes, 0) / NVTEMPERATURE_DIVISOR;
                currentIndex += TEMPERATURE_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, SLOPE_OFFSET_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LineOffset2 = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += SLOPE_OFFSET_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, TEMPERATURE_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.Temperature3 = (double)BitConverter.ToInt16(twoBytes, 0) / NVTEMPERATURE_DIVISOR;
                currentIndex += TEMPERATURE_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, SLOPE_OFFSET_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LineOffset3 = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += SLOPE_OFFSET_SIZE;

                MergeLineOffsetList.Add(mergeLineOffsetClass);
            }

            FactoryMergeLineOffsetList = new List<MergeLineOffsetClass>();
            for (int i = 0; i < NVNumberOfFactoryMergeLineOffsets; i++)
            {
                MergeLineOffsetClass mergeLineOffsetClass = new MergeLineOffsetClass();

                mergeLineOffsetClass.BlackLevelHigh = BlackLevelHigh;
                mergeLineOffsetClass.BlackLevelLow = BlackLevelLow;

                Array.Copy(data, currentIndex, twoBytes, 0, GAIN_INDEX_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.HighGainIndex = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += GAIN_INDEX_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, GAIN_INDEX_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LowGainIndex = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += GAIN_INDEX_SIZE;

                Array.Copy(data, currentIndex, fourBytes, 0, GAIN_RATIO_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(fourBytes);
                }
                mergeLineOffsetClass.GainRatio = (double)BitConverter.ToUInt32(fourBytes, 0) / NVGAIN_RATIO_DIVISOR;
                currentIndex += GAIN_RATIO_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, TEMPERATURE_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.Temperature1 = (double)BitConverter.ToInt16(twoBytes, 0) / NVTEMPERATURE_DIVISOR;
                currentIndex += TEMPERATURE_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, SLOPE_OFFSET_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LineOffset1 = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += SLOPE_OFFSET_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, TEMPERATURE_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.Temperature2 = (double)BitConverter.ToInt16(twoBytes, 0) / NVTEMPERATURE_DIVISOR;
                currentIndex += TEMPERATURE_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, SLOPE_OFFSET_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LineOffset2 = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += SLOPE_OFFSET_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, TEMPERATURE_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.Temperature3 = (double)BitConverter.ToInt16(twoBytes, 0) / NVTEMPERATURE_DIVISOR;
                currentIndex += TEMPERATURE_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, SLOPE_OFFSET_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LineOffset3 = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += SLOPE_OFFSET_SIZE;

                FactoryMergeLineOffsetList.Add(mergeLineOffsetClass);
            }
        }

        private void ParseNVStorageVersion3(byte[] data, int currentIndex)
        {
            byte[] twoBytes = new byte[2];
            byte[] fourBytes = new byte[4];

            Array.Copy(data, currentIndex, fourBytes, 0, ADUS_PER_ELECTRON_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            NVADUsPerElectron = (double)BitConverter.ToUInt32(fourBytes, 0) / NVGAIN_RATIO_DIVISOR;
            currentIndex += ADUS_PER_ELECTRON_SIZE;

            Array.Copy(data, currentIndex, fourBytes, 0, ADUS_PER_ELECTRON_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            NVFactoryADUsPerElectron = (double)BitConverter.ToUInt32(fourBytes, 0) / NVGAIN_RATIO_DIVISOR;
            currentIndex += ADUS_PER_ELECTRON_SIZE;

            Array.Copy(data, currentIndex, fourBytes, 0, GAIN_RATIO_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            NVGainRatio = (double)BitConverter.ToUInt32(fourBytes, 0) / NVGAIN_RATIO_DIVISOR;
            currentIndex += GAIN_RATIO_SIZE;

            Array.Copy(data, currentIndex, fourBytes, 0, GAIN_RATIO_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            NVFactoryGainRatio = (double)BitConverter.ToUInt32(fourBytes, 0) / NVGAIN_RATIO_DIVISOR;
            currentIndex += GAIN_RATIO_SIZE;

            Array.Copy(data, currentIndex, twoBytes, 0, NUMBER_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            NVNumberOfElectronsPerCountHighGains = BitConverter.ToInt16(twoBytes, 0);
            currentIndex += NUMBER_SIZE;

            Array.Copy(data, currentIndex, twoBytes, 0, NUMBER_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            NVNumberOfElectronsPerCountLowGains = BitConverter.ToInt16(twoBytes, 0);
            currentIndex += NUMBER_SIZE;

            Array.Copy(data, currentIndex, twoBytes, 0, NUMBER_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            NVNumberOfMergeLineOffsets = BitConverter.ToInt16(twoBytes, 0);
            currentIndex += NUMBER_SIZE;

            Array.Copy(data, currentIndex, twoBytes, 0, NUMBER_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            NVNumberOfFactoryMergeLineOffsets = BitConverter.ToInt16(twoBytes, 0);
            currentIndex += NUMBER_SIZE;

            FactoryElectronsPerCountHighGainList = new List<ElectronsPerCountClass>();
            for (int i = 0; i < NVNumberOfElectronsPerCountHighGains; i++)
            {
                ElectronsPerCountClass electronsPerCountClass = new ElectronsPerCountClass();

                Array.Copy(data, currentIndex, twoBytes, 0, GAIN_INDEX_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                electronsPerCountClass.Index = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += GAIN_INDEX_SIZE;

                Array.Copy(data, currentIndex, fourBytes, 0, COUNTS_PER_ELECTRON_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(fourBytes);
                }
                electronsPerCountClass.ElectronsPerCount = (double)BitConverter.ToUInt32(fourBytes, 0) / NVCOUNTS_PER_ELECTRON_DIVISOR;
                currentIndex += COUNTS_PER_ELECTRON_SIZE;

                FactoryElectronsPerCountHighGainList.Add(electronsPerCountClass);
            }

            FactoryElectronsPerCountLowGainList = new List<ElectronsPerCountClass>();
            for (int i = 0; i < NVNumberOfElectronsPerCountLowGains; i++)
            {
                ElectronsPerCountClass electronsPerCountClass = new ElectronsPerCountClass();

                Array.Copy(data, currentIndex, twoBytes, 0, GAIN_INDEX_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                electronsPerCountClass.Index = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += GAIN_INDEX_SIZE;

                Array.Copy(data, currentIndex, fourBytes, 0, COUNTS_PER_ELECTRON_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(fourBytes);
                }
                electronsPerCountClass.ElectronsPerCount = (double)BitConverter.ToUInt32(fourBytes, 0) / NVCOUNTS_PER_ELECTRON_DIVISOR;
                currentIndex += COUNTS_PER_ELECTRON_SIZE;

                FactoryElectronsPerCountLowGainList.Add(electronsPerCountClass);
            }

            MergeLineOffsetList = new List<MergeLineOffsetClass>();
            for (int i = 0; i < NVNumberOfMergeLineOffsets; i++)
            {
                MergeLineOffsetClass mergeLineOffsetClass = new MergeLineOffsetClass();

                Array.Copy(data, currentIndex, twoBytes, 0, BLACK_LEVEL_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.BlackLevelHigh = BitConverter.ToUInt16(twoBytes, 0);
                currentIndex += BLACK_LEVEL_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, BLACK_LEVEL_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.BlackLevelLow = BitConverter.ToUInt16(twoBytes, 0);
                currentIndex += BLACK_LEVEL_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, GAIN_INDEX_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.HighGainIndex = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += GAIN_INDEX_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, GAIN_INDEX_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LowGainIndex = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += GAIN_INDEX_SIZE;

                Array.Copy(data, currentIndex, fourBytes, 0, GAIN_RATIO_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(fourBytes);
                }
                mergeLineOffsetClass.GainRatio = (double)BitConverter.ToUInt32(fourBytes, 0) / NVGAIN_RATIO_DIVISOR;
                currentIndex += GAIN_RATIO_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, TEMPERATURE_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.Temperature1 = (double)BitConverter.ToInt16(twoBytes, 0) / NVTEMPERATURE_DIVISOR;
                currentIndex += TEMPERATURE_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, SLOPE_OFFSET_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LineOffset1 = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += SLOPE_OFFSET_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, TEMPERATURE_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.Temperature2 = (double)BitConverter.ToInt16(twoBytes, 0) / NVTEMPERATURE_DIVISOR;
                currentIndex += TEMPERATURE_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, SLOPE_OFFSET_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LineOffset2 = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += SLOPE_OFFSET_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, TEMPERATURE_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.Temperature3 = (double)BitConverter.ToInt16(twoBytes, 0) / NVTEMPERATURE_DIVISOR;
                currentIndex += TEMPERATURE_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, SLOPE_OFFSET_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LineOffset3 = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += SLOPE_OFFSET_SIZE;

                MergeLineOffsetList.Add(mergeLineOffsetClass);
            }

            FactoryMergeLineOffsetList = new List<MergeLineOffsetClass>();
            for (int i = 0; i < NVNumberOfFactoryMergeLineOffsets; i++)
            {
                MergeLineOffsetClass mergeLineOffsetClass = new MergeLineOffsetClass();

                Array.Copy(data, currentIndex, twoBytes, 0, BLACK_LEVEL_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.BlackLevelHigh = BitConverter.ToUInt16(twoBytes, 0);
                currentIndex += BLACK_LEVEL_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, BLACK_LEVEL_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.BlackLevelLow = BitConverter.ToUInt16(twoBytes, 0);
                currentIndex += BLACK_LEVEL_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, GAIN_INDEX_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.HighGainIndex = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += GAIN_INDEX_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, GAIN_INDEX_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LowGainIndex = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += GAIN_INDEX_SIZE;

                Array.Copy(data, currentIndex, fourBytes, 0, GAIN_RATIO_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(fourBytes);
                }
                mergeLineOffsetClass.GainRatio = (double)BitConverter.ToUInt32(fourBytes, 0) / NVGAIN_RATIO_DIVISOR;
                currentIndex += GAIN_RATIO_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, TEMPERATURE_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.Temperature1 = (double)BitConverter.ToInt16(twoBytes, 0) / NVTEMPERATURE_DIVISOR;
                currentIndex += TEMPERATURE_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, SLOPE_OFFSET_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LineOffset1 = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += SLOPE_OFFSET_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, TEMPERATURE_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.Temperature2 = (double)BitConverter.ToInt16(twoBytes, 0) / NVTEMPERATURE_DIVISOR;
                currentIndex += TEMPERATURE_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, SLOPE_OFFSET_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LineOffset2 = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += SLOPE_OFFSET_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, TEMPERATURE_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.Temperature3 = (double)BitConverter.ToInt16(twoBytes, 0) / NVTEMPERATURE_DIVISOR;
                currentIndex += TEMPERATURE_SIZE;

                Array.Copy(data, currentIndex, twoBytes, 0, SLOPE_OFFSET_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                mergeLineOffsetClass.LineOffset3 = BitConverter.ToInt16(twoBytes, 0);
                currentIndex += SLOPE_OFFSET_SIZE;

                FactoryMergeLineOffsetList.Add(mergeLineOffsetClass);
            }
        }

        private void StoreValue(object value, byte[] metaData, ref int arrayIndex, int size)
        {
            byte[] dataArray = null;

            switch (size)
            {
                case 2:
                    short targetShort = Convert.ToInt16(value);
                    dataArray = BitConverter.GetBytes(targetShort);
                    break;
                case 4:
                    int targetInt = Convert.ToInt32(value);
                    dataArray = BitConverter.GetBytes(targetInt);
                    break;
                case 8:
                    long targetLong = Convert.ToInt64(value);
                    dataArray = BitConverter.GetBytes(targetLong);
                    break;
            }

            if (dataArray != null)
            {
                Array.Reverse(dataArray);
                Array.Copy(dataArray, 0, metaData, arrayIndex, size);
                arrayIndex += size;
            }
        }

        private void UpdateElectronsPerCountList(List<ElectronsPerCountClass> electronsPerCountGainList, int gainIndex, double activeElectronsPerCountHigh)
        {
            ElectronsPerCountClass electronsPerCountClass = new ElectronsPerCountClass(gainIndex, activeElectronsPerCountHigh);

            for (int index = 0; index < electronsPerCountGainList.Count; index++)
            {
                if (electronsPerCountGainList[index].Index == gainIndex)
                {
                    // Replace the current entry
                    electronsPerCountGainList.RemoveAt(index);
                    electronsPerCountGainList.Insert(index, electronsPerCountClass);
                    return;
                }
            }
            electronsPerCountGainList.Add(electronsPerCountClass);
        }

        private void UpdateMergeLineOffsetList(List<MergeLineOffsetClass> mergeLineOffsetList, uint blackLevelHigh, uint blackLevelLow, int highGainIndex, int lowGainIndex,
            double gainRatio, double temperature1, int lineOffset1, double temperature2, int lineOffset2, double temperature3, int lineOffset3)
        {
            MergeLineOffsetClass mergeLineOffsetClass = new MergeLineOffsetClass(blackLevelHigh, blackLevelLow, highGainIndex, lowGainIndex, gainRatio, temperature1,
                lineOffset1, temperature2, lineOffset2, temperature3, lineOffset3);

            for (int index = 0; index < mergeLineOffsetList.Count; index++)
            {
                if ((mergeLineOffsetList[index].HighGainIndex == highGainIndex) && (mergeLineOffsetList[index].LowGainIndex == lowGainIndex))
                {
                    // Replace the current entry
                    mergeLineOffsetList.RemoveAt(index);
                    mergeLineOffsetList.Insert(index, mergeLineOffsetClass);
                    return;
                }
            }
            mergeLineOffsetList.Add(mergeLineOffsetClass);
        }

        #endregion Protected Methods

    }
}
