﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using FliUtils.Jobs;
using System.Windows;
using System.IO;
using System.Diagnostics;
using System.Windows.Interop;
using System.Windows.Forms;
using System.Windows.Media.Imaging;
using System.Drawing;
using System.Drawing.Imaging;
using System.Windows.Media;
using FliProCameraLib;

namespace FliProCameraLib
{
    /// <summary>
    /// This file contains properties and methods that are used to write files to disc
    /// </summary>
    public partial class FliProCamera : FliPro
    {
        //**************************************************************
        #region Constants and Enums

        #endregion Constants and Enums

        //**************************************************************
        #region Protected Properties

        protected object ImageWriteLock;

        #endregion Protected Properties

        //**************************************************************
        #region Public Properties

        #endregion Public Properties

        //**************************************************************
        #region Public Classes

        public class WriteFileSettingsClass
        {
            public WriteFileSettingsClass() { }

            public bool IsSaveExposureToFITFileChecked { get; set; }
            public bool IsSaveExposureToJPEGFileChecked { get; set; }
            public bool IsSaveExposureToPNGFileChecked { get; set; }
            public bool IsSaveExposureToBINFileChecked { get; set; }
            public bool IsSaveExposureToRCDFileChecked { get; set; }
            public bool IsSaveExposureToRAWFileChecked { get; set; }
            public bool IsSaveExposureToTIFFFileChecked { get; set; }
            public bool IsSaveMetaDataToFileChecked { get; set; }
            public bool SaveSectorsAsFiles { get; set; }

            public bool UseLowGainImageChecked { get; set; }
            public bool UseHighGainImageChecked { get; set; }
            public bool UseMergedImageChecked { get; set; }

            public string FileNameBase { get; set; }
            public string Folder { get; set; }
            public string NameHint { get; set; }

            public FliFrame Frame { get; set; }

        }

        #endregion Public Classes

        //**************************************************************
        #region Observable Properties
        #endregion Observable Properties

        //**************************************************************
        #region Public Collections

        /// <summary>
        /// The list of files to be written to the disk.
        /// </summary>
        public List<WriteFileSettingsClass> WriteFileList { get; set; }

        #endregion Public Collections

        //**************************************************************
        #region Public Methods

        public void WriteFile(FliFrame frame, string folder, string NameHint = "", bool forceFileWrite = false)
        {
            // Don't allow writing frame to file when this camera is already from a file
            if (((Type == CameraType.FILE) || (Type == CameraType.GENERATED)) && !forceFileWrite)
            {
                return;
            }

            if (!String.IsNullOrEmpty(folder))
            {
                if (fliProCameraManager == null)
                {
                    fliProCameraManager = FliProCameraManager.Instance;
                }
                string fileNameBase = fliProCameraManager.CalculateNameString(frame, NameHint);
                WriteFileSettingsClass writeFileClass = new WriteFileSettingsClass();

                writeFileClass.IsSaveExposureToFITFileChecked = IsSaveExposureToFITFileChecked;
                writeFileClass.IsSaveExposureToJPEGFileChecked = IsSaveExposureToJPEGFileChecked;
                writeFileClass.IsSaveExposureToPNGFileChecked = IsSaveExposureToPNGFileChecked;
                writeFileClass.IsSaveExposureToBINFileChecked = IsSaveExposureToBINFileChecked;
                writeFileClass.IsSaveExposureToRCDFileChecked = IsSaveExposureToRCDFileChecked;
                writeFileClass.IsSaveExposureToRAWFileChecked = IsSaveExposureToRAWFileChecked;
                writeFileClass.IsSaveExposureToTIFFFileChecked = IsSaveExposureToTIFFFileChecked;
                writeFileClass.IsSaveMetaDataToFileChecked = IsSaveMetaDataToFileChecked;
                writeFileClass.SaveSectorsAsFiles = SaveSectorsAsFiles;

                writeFileClass.UseLowGainImageChecked = UseLowGainImageChecked;
                writeFileClass.UseHighGainImageChecked = UseHighGainImageChecked;
                writeFileClass.UseMergedImageChecked = UseMergedImageChecked;

                writeFileClass.FileNameBase = fileNameBase;
                writeFileClass.Folder = folder;
                writeFileClass.NameHint = NameHint;

                writeFileClass.Frame = frame;

                InsertWriteFile(writeFileClass);
            }
        }

        public void WriteFileInvalidMetadata(FliFrame frame, string folder, string NameHint = "", bool forceFileWrite = false)
        {
            // Don't allow writing frame to file when this camera is already from a file
            if ((Type == CameraType.FILE) && !forceFileWrite)
            {
                return;
            }

            if (!String.IsNullOrEmpty(folder))
            {
                if (fliProCameraManager == null)
                {
                    fliProCameraManager = FliProCameraManager.Instance;
                }
                string fileNameBase = fliProCameraManager.CalculateNameString(frame);
                string fileName = String.Empty;
                List<string> nameList = new List<string>();

                // The 16 bit arrays have not been generated so only save the raw image data 
                fileName = GetValidFilename(folder, fileNameBase + "RawData_", FILE_EXTENSION_BIN);
                nameList.Add(fileName);
                SaveBINFile(frame, fileName, File_Type.RawData);

                if (IsSaveMetaDataToFileChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Metadata_", FILE_EXTENSION_TXT);
                    SaveMetadataFile(frame, fileName);
                }
            }
        }

        public void WriteFileSkipped(FliFrame frame, string folder, string NameHint = "", bool forceFileWrite = false)
        {
            // Don't allow writing frame to file when this camera is already from a file
            if ((Type == CameraType.FILE) && !forceFileWrite)
            {
                return;
            }

            if (!String.IsNullOrEmpty(folder))
            {
                if (frame.FrameMetaData == null)
                {
                    if (frame.RawData != null)
                    {
                        // Strip off the MetaData and put it into the RawMetaData property
                        frame.RawMetaData = new byte[FliProFrameMetadata.GetMetadataLengthFromArray(frame.RawData)];
                        Buffer.BlockCopy(frame.RawData, 0, frame.RawMetaData, 0, frame.RawMetaData.Length);

                        // Parse the raw bytes into the metadata structure
                        frame.FrameMetaData = new FliProFrameMetadata(frame.RawMetaData);
                        frame.FrameMetaData.TestLowGainAbsolute = GetGainValueAbsoluteFromIndex(LowGainTable, frame.FrameMetaData.LowGain);
                        frame.FrameMetaData.TestHighGainAbsolute = GetGainValueAbsoluteFromIndex(HighGainTable, frame.FrameMetaData.HighGain);

                        // Temporary code
                        frame.FrameMetaData.NumberOfDataChannels = 8;
                        //FrameMetaData.PixelOrderedImage = !UnscrambleImage;

                        frame.IsValid = frame.IsValid && (frame.FrameMetaData.MagicNumber == FliProFrameMetadata.MAGIC_NUMBER);
                        if (!frame.IsValid)
                        {
                            // This is corrupted data so we need to just abort and return
                            LogEvent("ParseRawImageData: Frame data invalid");
                            CancelExposure();
                            WriteFileInvalidMetadata(frame, fliProCameraManager.RawFilePath, "BadMetadataFile");
                            return;
                        }

                        BitsPerPixel = frame.FrameMetaData.DataPixelBitDepth;

                        // Insure that the FrameMetaData.StartingExposureRow is always even. If it is odd we will crash with Index out of range while parsing.
                        // This is a problem in the image generation code that once fixed should never occur.
                        if (frame.FrameMetaData.StartingExposureRow % 2 > 0)
                        {
                            frame.FrameMetaData.StartingExposureRow--;
                        }

                        int totalHeight = frame.FrameMetaData.PreDummyRows + (frame.FrameMetaData.ImageStopRow - frame.FrameMetaData.ImageStartRow + 1) + frame.FrameMetaData.PostDummyRows;
                        frame.FrameAreaWH = new FliPro.FliRectangleWH(0, 0,
                            (frame.FrameMetaData.ImageStopColumn - frame.FrameMetaData.ImageStartColumn + 1),
                            (frame.FrameMetaData.ImageStopRow - frame.FrameMetaData.ImageStartRow + 1));
                        frame.FrameAreaWHTotal = new FliPro.FliRectangleWH(0, 0, (frame.FrameMetaData.ImageStopColumn - frame.FrameMetaData.ImageStartColumn + 1) +
                            frame.FrameMetaData.DummyPixelsPerRow, totalHeight);
                    }
                }
                //string frameNumberString = (frame.FrameMetaData == null ? "0000" : frame.FrameMetaData.FrameNumber.ToString("0000")) + "_";
                //string targetSerialNumber = SerialNumber;
                //if (SerialNumber.StartsWith("Camera_"))
                //{
                //    // This is an image from a file so filter out the extra substrings
                //    targetSerialNumber = targetSerialNumber.Substring(7);
                //    int index = targetSerialNumber.IndexOf("_");
                //    targetSerialNumber = targetSerialNumber.Substring(0, index);
                //}
                string fileNameBase = fliProCameraManager.CalculateNameString(frame);
                string fileName = String.Empty;
                List<string> nameList = new List<string>();

                // The 16 bit arrays have not been generated so only save the raw image data 
                fileName = GetValidFilename(folder, fileNameBase + "RawData_", FILE_EXTENSION_BIN);
                nameList.Add(fileName);
                SaveBINFile(frame, fileName, File_Type.RawData);

                if (IsSaveMetaDataToFileChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Metadata_", FILE_EXTENSION_TXT);
                    SaveMetadataFile(frame, fileName);
                }
            }
        }

        public void WriteFileWithFileName(FliFrame frame, string folder, string fileName, string extension)
        {
            string targetSerialNumber = SerialNumber;
            string fileNameBase = fileName + "_" + DateTime.Now.ToString("yyyy_MM_dd") + "_";
            string completeFileName = String.Empty;
            BitmapSource sourceLow = null;
            BitmapSource sourceHigh = null;
            BitmapSource sourceMerge = null;
            List<string> nameList = new List<string>();

            if (IsSaveExposureToFITFileChecked || extension.Equals(FILE_EXTENSION_FIT, StringComparison.OrdinalIgnoreCase))
            {
                if (UseLowGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Low_", FILE_EXTENSION_FIT);
                    nameList.Add(fileName);
                    FITSImageProcessing.SaveFitsFile(fileName, frame, FliFrame.FliImageType.LowGainDummy);
                }
                if (UseHighGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "High_", FILE_EXTENSION_FIT);
                    nameList.Add(fileName);
                    FITSImageProcessing.SaveFitsFile(fileName, frame, FliFrame.FliImageType.HighGainDummy);
                }
                if (UseMergedImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Merge_", FILE_EXTENSION_FIT);
                    nameList.Add(fileName);
                    FITSImageProcessing.SaveFitsFile(fileName, frame, FliFrame.FliImageType.MergedDummy);
                }
            }

            if (IsSaveExposureToJPEGFileChecked || extension.Equals(FILE_EXTENSION_JPEG, StringComparison.OrdinalIgnoreCase))
            {
                if (UseLowGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Low_", FILE_EXTENSION_JPEG);
                    nameList.Add(fileName);
                    sourceLow = SaveImageFile8Bit(frame, fileName, ImageFormat.Jpeg, sourceLow, frame.GetTransformedPixelArray(FliFrame.FliImageType.LowGain));
                }
                if (UseHighGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "High_", FILE_EXTENSION_JPEG);
                    sourceHigh = SaveImageFile8Bit(frame, fileName, ImageFormat.Jpeg, sourceHigh, frame.GetTransformedPixelArray(FliFrame.FliImageType.HighGain));
                }
                if (UseMergedImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Merge_", FILE_EXTENSION_JPEG);
                    sourceMerge = SaveImageFile8Bit(frame, fileName, ImageFormat.Jpeg, sourceMerge, frame.GetTransformedPixelArray(FliFrame.FliImageType.Merged));
                }
            }

            if (IsSaveExposureToPNGFileChecked || extension.Equals(FILE_EXTENSION_PNG, StringComparison.OrdinalIgnoreCase))
            {
                if (UseLowGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Low_", FILE_EXTENSION_PNG);
                    nameList.Add(fileName);
                    sourceLow = SaveImageFile(frame, fileName, ImageFormat.Png, sourceLow, frame.GetPixelArray(FliFrame.FliImageType.LowGain));
                }
                if (UseHighGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "High_", FILE_EXTENSION_PNG);
                    sourceHigh = SaveImageFile(frame, fileName, ImageFormat.Png, sourceHigh, frame.GetPixelArray(FliFrame.FliImageType.HighGain));
                }
                if (UseMergedImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Merge_", FILE_EXTENSION_PNG);
                    sourceMerge = SaveImageFile(frame, fileName, ImageFormat.Png, sourceMerge, frame.GetPixelArray(FliFrame.FliImageType.Merged));
                }
            }

            if (IsSaveExposureToBINFileChecked || extension.Equals(FILE_EXTENSION_BIN, StringComparison.OrdinalIgnoreCase))
            {
                // Then first the low only as a 16 bit file 
                if (UseLowGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Low_", FILE_EXTENSION_BIN);
                    nameList.Add(fileName);
                    SaveBINFile(frame, fileName, File_Type.Low);
                }
                if (UseHighGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "High_", FILE_EXTENSION_BIN);
                    SaveBINFile(frame, fileName, File_Type.High);
                }
                if (UseMergedImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Merge_", FILE_EXTENSION_BIN);
                    SaveBINFile(frame, fileName, File_Type.Merged);
                }
            }

            if (IsSaveExposureToRCDFileChecked || extension.Equals(FILE_EXTENSION_RCD, StringComparison.OrdinalIgnoreCase))
            {
                // First save the Combined file which is exactly how we received it from the board.
                fileName = GetValidFilename(folder, fileNameBase + "RCDData_", FILE_EXTENSION_RCD);
                nameList.Add(fileName);
                SaveBINFile(frame, fileName, File_Type.RawData);

            }

            if (IsSaveExposureToRAWFileChecked || extension.Equals(FILE_EXTENSION_RAW))
            {
                if (UseLowGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Low_", FILE_EXTENSION_RAW);
                    nameList.Add(fileName);
                    SaveRAWFile(frame.GetPixelArray(FliFrame.FliImageType.LowGain), fileName);
                }
                if (UseHighGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "High_", FILE_EXTENSION_RAW);
                    SaveRAWFile(frame.GetPixelArray(FliFrame.FliImageType.HighGain), fileName);
                }
                if (UseMergedImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Merge_", FILE_EXTENSION_RAW);
                    SaveRAWFile(frame.GetPixelArray(FliFrame.FliImageType.Merged), fileName);
                }
            }

            if (IsSaveExposureToTIFFFileChecked || extension.Equals(FILE_EXTENSION_TIFF, StringComparison.OrdinalIgnoreCase))
            {
                if (UseLowGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Low_", FILE_EXTENSION_TIFF);
                    nameList.Add(fileName);
                    sourceLow = SaveImageFile(frame, fileName, ImageFormat.Tiff, sourceLow, frame.GetPixelArray(FliFrame.FliImageType.LowGain));
                }
                if (UseHighGainImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "High_", FILE_EXTENSION_TIFF);
                    sourceHigh = SaveImageFile(frame, fileName, ImageFormat.Tiff, sourceHigh, frame.GetPixelArray(FliFrame.FliImageType.HighGain));
                }
                if (UseMergedImageChecked)
                {
                    fileName = GetValidFilename(folder, fileNameBase + "Merge_", FILE_EXTENSION_TIFF);
                    sourceMerge = SaveImageFile(frame, fileName, ImageFormat.Tiff, sourceMerge, frame.GetPixelArray(FliFrame.FliImageType.Merged));
                }
            }

            if (IsSaveMetaDataToFileChecked)
            {
                fileName = GetValidFilename(folder, fileNameBase + "Metadata_", FILE_EXTENSION_TXT);
                SaveMetadataFile(frame, fileName);
            }

            if (SaveSectorsAsFiles)
            {
                WriteFileSectors(frame, folder, nameList);
            }
        }

        public bool WriteFile(FliFrame.FliImageType imageType, string fileName, FliFrame frame)
        {
            BitmapSource source = null;
            bool fileExtensionValid = false;

            if (!String.IsNullOrEmpty(fileName))
            {
                if (fileName.EndsWith(FILE_EXTENSION_FIT))
                {
                    FITSImageProcessing.SaveFitsFile(fileName, frame, imageType);
                    fileExtensionValid = true;
                }

                if (fileName.EndsWith(FILE_EXTENSION_JPEG))
                {
                    source = SaveImageFile8Bit(frame, fileName, ImageFormat.Jpeg, source, frame.GetTransformedPixelArray(imageType));
                    fileExtensionValid = true;
                }

                if (fileName.EndsWith(FILE_EXTENSION_PNG))
                {
                    source = SaveImageFile(frame, fileName, ImageFormat.Png, source, frame.GetPixelArray(imageType));
                    fileExtensionValid = true;
                }

                if (fileName.EndsWith(FILE_EXTENSION_RAW))
                {
                    SaveRAWFile(frame.GetPixelArray(imageType), fileName);
                    fileExtensionValid = true;
                }

                if (fileName.EndsWith(FILE_EXTENSION_RCD))
                {
                    SaveBINFile(frame, fileName, File_Type.RawData);
                    fileExtensionValid = true;
                }

                if (fileName.EndsWith(FILE_EXTENSION_BIN))
                {
                    File_Type fileType = File_Type.RawData;
                    if ((imageType == FliFrame.FliImageType.HighGain) || (imageType == FliFrame.FliImageType.HighGainDummy))
                    {
                        fileType = File_Type.High;
                    }
                    else if ((imageType == FliFrame.FliImageType.LowGain) || (imageType == FliFrame.FliImageType.LowGainDummy))
                    {
                        fileType = File_Type.Low;
                    }
                    else
                    {
                        fileType = File_Type.Merged;
                    }
                    SaveBINFile(frame, fileName, fileType);
                    fileExtensionValid = true;
                }

                if (fileName.EndsWith(FILE_EXTENSION_TIFF))
                {
                    source = SaveImageFile(frame, fileName, ImageFormat.Tiff, source, frame.GetPixelArray(imageType));
                    fileExtensionValid = true;
                }
            }
            return fileExtensionValid;
        }

        private void WriteFileSectors(FliFrame frame, string folder, List<string> nameList)
        {
            FliFrame subframe = new FliFrame(frame.Camera);
            ushort[,] newLowArray = new ushort[frame.FrameAreaWH.H, frame.FrameAreaWH.W / 8];
            string fileName = String.Empty;
            BitmapSource source = null;
            int fileNameIndex = 0;
            int index = 0;

            subframe.FrameMetaData = frame.FrameMetaData;
            subframe.FrameAreaWH = new FliPro.FliRectangleWH(0, 0, frame.FrameAreaWH.W / 8, frame.FrameAreaWH.H);
            subframe.FrameMetaData.DataPixelBitDepth = FliFrame.BITS_PER_PIXEL_16;
            subframe.FrameMetaData.SensorPixelBitDepth = FliFrame.BITS_PER_PIXEL_16;
            subframe.FrameMetaData.HorizontalPixels = subframe.FrameAreaWH.W;
            subframe.FrameMetaData.ImageStartColumn = 255;
            subframe.FrameMetaData.PixelOrderedImage = true;
            subframe.FrameMetaData.HorizontalBinning = 1;
            subframe.FrameMetaData.VerticalBinning = 1;
            subframe.FrameMetaData.HighGain = fliProCameraManager.ActiveCamera.GetMetadataIndexFromGainIndex(
                        fliProCameraManager.ActiveCamera.HighGainTable, fliProCameraManager.ActiveCamera.HighGainIndex);
            subframe.FrameMetaData.LowGain = fliProCameraManager.ActiveCamera.GetMetadataIndexFromGainIndex(
                        fliProCameraManager.ActiveCamera.LowGainTable, fliProCameraManager.ActiveCamera.LowGainIndex);

            for (int i = 0; i < 8; i++)
            {
                int offset = i * subframe.FrameAreaWH.W;
                for (int y = 0; y < frame.FrameAreaWH.H; y++)
                {
                    for (int x = 0; x < subframe.FrameAreaWH.W; x++)
                    {
                        newLowArray[y, x] = frame.GetPixel(FliFrame.FliImageType.LowGainDummy, offset + x, y);
                    }
                }
                BitsPerPixel = FliFrame.BITS_PER_PIXEL_16;
                subframe.LowGainDummyPixelArray = subframe.CopyToOneDimensionalArray(newLowArray);
                fileNameIndex = 0;  // Reset the index

                // Now write the files
                if (IsSaveExposureToFITFileChecked)
                {
                    fileName = nameList[fileNameIndex];
                    index = fileName.LastIndexOf(".");
                    fileName = fileName.Substring(0, index) + "-Sector" + i.ToString() + fileName.Substring(index);
                    fileNameIndex++;
                    FITSImageProcessing.SaveFitsFile(fileName, subframe, FliFrame.FliImageType.LowGainDummy);
                }

                if (IsSaveExposureToJPEGFileChecked)
                {
                    fileName = nameList[fileNameIndex];
                    index = fileName.LastIndexOf(".");
                    fileName = fileName.Substring(0, index) + "-Sector" + i.ToString() + fileName.Substring(index);
                    fileNameIndex++;
                    source = SaveImageFile8Bit(subframe, fileName, ImageFormat.Jpeg, source, subframe.GetTransformedPixelArray(FliFrame.FliImageType.LowGain));
                }

                if (IsSaveExposureToPNGFileChecked)
                {
                    fileName = nameList[fileNameIndex];
                    index = fileName.LastIndexOf(".");
                    fileName = fileName.Substring(0, index) + "-Sector" + i.ToString() + fileName.Substring(index);
                    fileNameIndex++;
                    source = SaveImageFile(subframe, fileName, ImageFormat.Png, source, subframe.GetPixelArray(FliFrame.FliImageType.LowGain));
                }

                if (IsSaveExposureToBINFileChecked)
                {
                    fileName = nameList[fileNameIndex];
                    index = fileName.LastIndexOf(".");
                    fileName = fileName.Substring(0, index) + "-Sector" + i.ToString() + fileName.Substring(index);
                    fileNameIndex++;
                    SaveBINFile(subframe, fileName, File_Type.Low);
                }

                if (IsSaveExposureToRCDFileChecked)
                {
                    fileName = nameList[fileNameIndex];
                    index = fileName.LastIndexOf(".");
                    fileName = fileName.Substring(0, index) + "-Sector" + i.ToString() + fileName.Substring(index);
                    fileNameIndex++;
                    SaveBINFile(subframe, fileName, File_Type.Low);
                }

                if (IsSaveExposureToRAWFileChecked)
                {
                    fileName = nameList[fileNameIndex];
                    index = fileName.LastIndexOf(".");
                    fileName = fileName.Substring(0, index) + "-Sector" + i.ToString() + fileName.Substring(index);
                    fileNameIndex++;
                    SaveRAWFile(subframe.GetPixelArray(FliFrame.FliImageType.LowGain), fileName);
                }

                if (IsSaveExposureToTIFFFileChecked)
                {
                    fileName = nameList[fileNameIndex];
                    index = fileName.LastIndexOf(".");
                    fileName = fileName.Substring(0, index) + "-Sector" + i.ToString() + fileName.Substring(index);
                    fileNameIndex++;
                    source = SaveImageFile(subframe, fileName, ImageFormat.Tiff, source, subframe.GetPixelArray(FliFrame.FliImageType.LowGain));
                }
            }
        }

        public void SaveBINFile(FliFrame frame, string fileName, File_Type fileType)
        {
            try
            {
                FileStream fs = new FileStream(fileName, FileMode.CreateNew);
                // Create the writer for data.
                BinaryWriter w = new BinaryWriter(fs);
                ushort[] pixelArray;
                bool tempHDRMode = false;
                bool tempPixelOrderedImage = false;
                int tempDataPixelBitDepth = 0;
                int tempSensorPixelBitDepth = 0;

                if ((frame.RawData == null) && ((fileType == File_Type.RawData) || (fileType == File_Type.Low)))
                {
                    // This is data from an old camera so write the MetaData and the low gain array
                    tempHDRMode = frame.FrameMetaData.HDRMode;
                    tempPixelOrderedImage = frame.FrameMetaData.PixelOrderedImage;
                    tempDataPixelBitDepth = frame.FrameMetaData.DataPixelBitDepth;
                    tempSensorPixelBitDepth = frame.FrameMetaData.SensorPixelBitDepth;
                    if (fileType == File_Type.Low)
                    {
                        frame.FrameMetaData.HDRMode = false;
                    }
                    frame.FrameMetaData.PixelOrderedImage = true;  // Need to set this flag so that the image is read back in properly
                    frame.FrameMetaData.SensorPixelBitDepth = FliFrame.BITS_PER_PIXEL_16;
                    frame.FrameMetaData.DataPixelBitDepth = FliFrame.BITS_PER_PIXEL_16;
                    w.Write(frame.FrameMetaData.GetByteArray());
                    frame.FrameMetaData.HDRMode = tempHDRMode;
                    frame.FrameMetaData.PixelOrderedImage = tempPixelOrderedImage;
                    frame.FrameMetaData.DataPixelBitDepth = tempDataPixelBitDepth;
                    frame.FrameMetaData.SensorPixelBitDepth = tempSensorPixelBitDepth;
                    pixelArray = frame.GetPixelArray(FliFrame.FliImageType.LowGainDummy);
                    byte[] byteArr = new byte[pixelArray.Length * sizeof(ushort)];
                    Buffer.BlockCopy(pixelArray, 0, byteArr, 0, byteArr.Length);
                    w.Write(byteArr);
                    pixelArray = frame.GetPixelArray(FliFrame.FliImageType.HighGainDummy);
                    if ((fileType != File_Type.Low) && (pixelArray != null))
                    {
                        byteArr = new byte[pixelArray.Length * sizeof(ushort)];
                        Buffer.BlockCopy(pixelArray, 0, byteArr, 0, byteArr.Length);
                        w.Write(byteArr);
                    }
                }
                else   // This is a new camera
                {
                    switch (fileType)
                    {
                        case File_Type.RawData:
                            if (frame.RawData != null)
                            {
                                w.Write(frame.RawData);
                            }
                            break;

                        case File_Type.High:
                            pixelArray = frame.GetPixelArray(FliFrame.FliImageType.HighGainDummy);
                            if (pixelArray != null)
                            {
                                tempHDRMode = frame.FrameMetaData.HDRMode;
                                tempPixelOrderedImage = frame.FrameMetaData.PixelOrderedImage;
                                tempDataPixelBitDepth = frame.FrameMetaData.DataPixelBitDepth;
                                tempSensorPixelBitDepth = frame.FrameMetaData.SensorPixelBitDepth;
                                frame.FrameMetaData.HDRMode = false;
                                frame.FrameMetaData.PixelOrderedImage = true;  // Need to set this flag so that the image is read back in properly
                                frame.FrameMetaData.SensorPixelBitDepth = FliFrame.BITS_PER_PIXEL_16;
                                frame.FrameMetaData.DataPixelBitDepth = FliFrame.BITS_PER_PIXEL_16;
                                w.Write(frame.FrameMetaData.GetByteArray());
                                frame.FrameMetaData.HDRMode = tempHDRMode;
                                frame.FrameMetaData.PixelOrderedImage = tempPixelOrderedImage;
                                frame.FrameMetaData.DataPixelBitDepth = tempDataPixelBitDepth;
                                frame.FrameMetaData.SensorPixelBitDepth = tempSensorPixelBitDepth;
                                byte[] byteArr = new byte[pixelArray.Length * sizeof(ushort)];
                                Buffer.BlockCopy(pixelArray, 0, byteArr, 0, byteArr.Length);
                                w.Write(byteArr);
                            }
                            break;

                        case File_Type.Low:
                            pixelArray = frame.GetPixelArray(FliFrame.FliImageType.LowGainDummy);
                            if (pixelArray != null)
                            {
                                tempHDRMode = frame.FrameMetaData.HDRMode;
                                tempPixelOrderedImage = frame.FrameMetaData.PixelOrderedImage;
                                tempDataPixelBitDepth = frame.FrameMetaData.DataPixelBitDepth;
                                tempSensorPixelBitDepth = frame.FrameMetaData.SensorPixelBitDepth;
                                frame.FrameMetaData.HDRMode = false;
                                frame.FrameMetaData.PixelOrderedImage = true;  // Need to set this flag so that the image is read back in properly
                                frame.FrameMetaData.SensorPixelBitDepth = FliFrame.BITS_PER_PIXEL_16;
                                frame.FrameMetaData.DataPixelBitDepth = FliFrame.BITS_PER_PIXEL_16;
                                w.Write(frame.FrameMetaData.GetByteArray());
                                frame.FrameMetaData.HDRMode = tempHDRMode;
                                frame.FrameMetaData.PixelOrderedImage = tempPixelOrderedImage;
                                frame.FrameMetaData.DataPixelBitDepth = tempDataPixelBitDepth;
                                frame.FrameMetaData.SensorPixelBitDepth = tempSensorPixelBitDepth;
                                byte[] byteArr = new byte[pixelArray.Length * sizeof(ushort)];
                                Buffer.BlockCopy(pixelArray, 0, byteArr, 0, byteArr.Length);
                                w.Write(byteArr);
                            }
                            break;

                        case File_Type.Merged:
                            pixelArray = frame.GetPixelArray(FliFrame.FliImageType.MergedDummy);
                            if (pixelArray != null)
                            {
                                tempHDRMode = frame.FrameMetaData.HDRMode;
                                tempPixelOrderedImage = frame.FrameMetaData.PixelOrderedImage;
                                tempDataPixelBitDepth = frame.FrameMetaData.DataPixelBitDepth;
                                tempSensorPixelBitDepth = frame.FrameMetaData.SensorPixelBitDepth;
                                frame.FrameMetaData.HDRMode = false;
                                frame.FrameMetaData.PixelOrderedImage = true;  // Need to set this flag so that the image is read back in properly
                                frame.FrameMetaData.SensorPixelBitDepth = FliFrame.BITS_PER_PIXEL_16;
                                frame.FrameMetaData.DataPixelBitDepth = FliFrame.BITS_PER_PIXEL_16;
                                w.Write(frame.FrameMetaData.GetByteArray());
                                frame.FrameMetaData.HDRMode = tempHDRMode;
                                frame.FrameMetaData.PixelOrderedImage = tempPixelOrderedImage;
                                frame.FrameMetaData.DataPixelBitDepth = tempDataPixelBitDepth;
                                frame.FrameMetaData.SensorPixelBitDepth = tempSensorPixelBitDepth;
                                byte[] byteArr = new byte[pixelArray.Length * sizeof(ushort)];
                                Buffer.BlockCopy(pixelArray, 0, byteArr, 0, byteArr.Length);
                                w.Write(byteArr);
                            }
                            break;
                    }
                }
                w.Close();
                fs.Close();
            }
            catch { }
        }

        /// <summary>
        /// Creates the target file type under the given filename and returns the generated BitmapSource for reuse
        /// if more than one file is requested.
        /// </summary>
        /// <param name="frame"></param>
        /// <param name="fileName"></param>
        /// <param name="imageFormat"></param>
        /// <param name="source"></param>
        /// <returns></returns>
        public BitmapSource SaveImageFile(FliFrame frame, string fileName, ImageFormat imageFormat, BitmapSource source, ushort[] targetArray)
        {
            // Create an Image from the byte array
            if (source == null)
            {
                if (targetArray != null)
                {
                    byte[] byteArr = new byte[targetArray.Length * 2];
                    Buffer.BlockCopy(targetArray, 0, byteArr, 0, byteArr.Length);
                    Bitmap bitmap = new Bitmap(frame.FrameAreaWHTotal.W, frame.FrameAreaWHTotal.H,
                        System.Drawing.Imaging.PixelFormat.Format16bppGrayScale);
                    var bitmap_data = bitmap.LockBits(new Rectangle(0, 0, bitmap.Width, bitmap.Height),
                        ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format16bppGrayScale);
                    Marshal.Copy(byteArr, 0, bitmap_data.Scan0, byteArr.Length);
                    bitmap.UnlockBits(bitmap_data);

                    Rectangle rect = new Rectangle(0, 0, frame.FrameAreaWHTotal.W, frame.FrameAreaWHTotal.H);
                    BitmapData bitmapData = bitmap.LockBits(rect, ImageLockMode.ReadOnly, bitmap.PixelFormat);

                    source = BitmapSource.Create(bitmap.Width,
                                                 bitmap.Height,
                                                 bitmap.HorizontalResolution,
                                                 bitmap.VerticalResolution,
                                                 PixelFormats.Gray16,
                                                 null,
                                                 bitmapData.Scan0,
                                                 bitmapData.Stride * bitmap.Height,
                                                 bitmapData.Stride);

                    bitmap.UnlockBits(bitmapData);
                }
            }

            if (source != null)
            {
                FileStream stream = new FileStream(fileName, FileMode.Create);

                if (imageFormat == ImageFormat.Tiff)
                {
                    TiffBitmapEncoder tiffEncoder = new TiffBitmapEncoder();

                    tiffEncoder.Compression = TiffCompressOption.Zip;
                    tiffEncoder.Frames.Add(BitmapFrame.Create(source));
                    tiffEncoder.Save(stream);
                }
                else if (imageFormat == ImageFormat.Jpeg)
                {
                    JpegBitmapEncoder jpegEncoder = new JpegBitmapEncoder();

                    jpegEncoder.QualityLevel = 100;  // Save 100% of the data
                    jpegEncoder.Frames.Add(BitmapFrame.Create(source));
                    jpegEncoder.Save(stream);
                }
                else if (imageFormat == ImageFormat.Png)
                {
                    PngBitmapEncoder pngEncoder = new PngBitmapEncoder();

                    pngEncoder.Frames.Add(BitmapFrame.Create(source));
                    pngEncoder.Save(stream);
                }

                stream.Close();
            }

            return source;
        }

        /// <summary>
        /// Creates the target file type under the given filename and returns the generated BitmapSource for reuse
        /// if more than one file is requested.
        /// </summary>
        /// <param name="frame"></param>
        /// <param name="fileName"></param>
        /// <param name="imageFormat"></param>
        /// <param name="source"></param>
        /// <returns></returns>
        public BitmapSource SaveImageFile8Bit(FliFrame frame, string fileName, ImageFormat imageFormat, BitmapSource source, byte[] targetArray)
        {
            // Create an Image from the byte array
            if (source == null)
            {
                if (targetArray != null)
                {
                    byte[] byteArr = new byte[targetArray.Length];
                    Buffer.BlockCopy(targetArray, 0, byteArr, 0, byteArr.Length);
                    Bitmap bitmap = new Bitmap(frame.FrameAreaWHTotal.W, frame.FrameAreaWHTotal.H,
                        System.Drawing.Imaging.PixelFormat.Format8bppIndexed);
                    var bitmap_data = bitmap.LockBits(new Rectangle(0, 0, bitmap.Width, bitmap.Height),
                        ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format8bppIndexed);
                    Marshal.Copy(byteArr, 0, bitmap_data.Scan0, byteArr.Length);
                    bitmap.UnlockBits(bitmap_data);

                    Rectangle rect = new Rectangle(0, 0, frame.FrameAreaWHTotal.W, frame.FrameAreaWHTotal.H);
                    BitmapData bitmapData = bitmap.LockBits(rect, ImageLockMode.ReadOnly, bitmap.PixelFormat);

                    source = BitmapSource.Create(bitmap.Width,
                                                 bitmap.Height,
                                                 bitmap.HorizontalResolution,
                                                 bitmap.VerticalResolution,
                                                 PixelFormats.Gray8,
                                                 null,
                                                 bitmapData.Scan0,
                                                 bitmapData.Stride * bitmap.Height,
                                                 bitmapData.Stride);

                    bitmap.UnlockBits(bitmapData);
                }
            }

            if (source != null)
            {
                FileStream stream = new FileStream(fileName, FileMode.Create);

                if (imageFormat == ImageFormat.Tiff)
                {
                    TiffBitmapEncoder tiffEncoder = new TiffBitmapEncoder();

                    tiffEncoder.Compression = TiffCompressOption.Zip;
                    tiffEncoder.Frames.Add(BitmapFrame.Create(source));
                    tiffEncoder.Save(stream);
                }
                else if (imageFormat == ImageFormat.Jpeg)
                {
                    JpegBitmapEncoder jpegEncoder = new JpegBitmapEncoder();

                    jpegEncoder.QualityLevel = 100;  // Save 100% of the data
                    jpegEncoder.Frames.Add(BitmapFrame.Create(source));
                    jpegEncoder.Save(stream);
                }
                else if (imageFormat == ImageFormat.Png)
                {
                    PngBitmapEncoder pngEncoder = new PngBitmapEncoder();

                    pngEncoder.Frames.Add(BitmapFrame.Create(source));
                    pngEncoder.Save(stream);
                }

                stream.Close();
            }

            return source;
        }

        /// <summary>
        /// Creates the JPeg file type under the given filename
        /// </summary>
        /// <param name="frame"></param>
        /// <param name="fileName"></param>
        /// <param name="imageFormat"></param>
        /// <param name="source"></param>
        /// <returns></returns>
        public void SaveJPEGFile(FliFrame frame, string fileName, ImageFormat imageFormat)
        {
            // Create an Image from the byte array
            BitmapSource source = null;
            ushort[] targetArray = frame.LowGainDummyPixelArray;

            if (UseMergedImageChecked && (frame.MergedDummyPixelArray != null))
            {
                targetArray = frame.MergedDummyPixelArray;
            }
            if (targetArray != null)
            {
                byte[] byteArr = new byte[targetArray.Length];
                for (int i = 0; i < targetArray.Length; i++)
                {
                    byteArr[i] = (byte)(targetArray[i] / 16);
                }
                Bitmap bitmap = new Bitmap(frame.FrameAreaWHTotal.W, frame.FrameAreaWHTotal.H);
                var bitmap_data = bitmap.LockBits(new Rectangle(0, 0, bitmap.Width, bitmap.Height),
                    ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format8bppIndexed);
                Marshal.Copy(byteArr, 0, bitmap_data.Scan0, byteArr.Length);
                bitmap.UnlockBits(bitmap_data);

                Rectangle rect = new Rectangle(0, 0, frame.FrameAreaWHTotal.W, frame.FrameAreaWHTotal.H);
                BitmapData bitmapData = bitmap.LockBits(rect, ImageLockMode.ReadOnly, bitmap.PixelFormat);

                source = BitmapSource.Create(bitmap.Width,
                                             bitmap.Height,
                                             bitmap.HorizontalResolution,
                                             bitmap.VerticalResolution,
                                             PixelFormats.Gray8,
                                             null,
                                             bitmapData.Scan0,
                                             bitmapData.Stride * bitmap.Height,
                                             bitmapData.Stride);

                bitmap.UnlockBits(bitmapData);
            }

            if (source != null)
            {
                FileStream stream = new FileStream(fileName, FileMode.Create);

                JpegBitmapEncoder jpegEncoder = new JpegBitmapEncoder();

                jpegEncoder.QualityLevel = 100;  // Save 100% of the data
                jpegEncoder.Frames.Add(BitmapFrame.Create(source));
                jpegEncoder.Save(stream);

                stream.Close();
            }
        }

        public void SaveMetadataFile(FliFrame frame, string fileName)
        {
            string formatedMetadata = frame.FrameMetaData.ToString();
            string formattedRawData = Environment.NewLine + "Raw Meta Data:" + Environment.NewLine + "[";

            if (frame.RawMetaData == null)
            {
                frame.RawMetaData = frame.FrameMetaData.GetByteArray();
            }

            for (int i = 0; i < frame.RawMetaData.Length; i++)
            {
                formattedRawData += frame.RawMetaData[i].ToString("X2");
                if (i < frame.RawMetaData.Length - 1)
                {
                    // Do this for all but the last one
                    if (((i + 1) % 20) == 0)
                    {
                        formattedRawData += "," + Environment.NewLine + " ";
                    }
                    else
                    {
                        formattedRawData += ", ";
                    }
                }
            }
            formattedRawData += "]";

            try
            {
                System.IO.File.WriteAllText(fileName, formatedMetadata + formattedRawData);
            }
            catch { }
        }

        public void SaveRAWFile(ushort[] targetArray, string fileName)
        {
            if (targetArray != null)
            {
                FileStream fs = new FileStream(fileName, FileMode.CreateNew);
                // Create the writer for data.
                BinaryWriter w = new BinaryWriter(fs);

                // Copy the 16 bit pixel data from the target gain array into a byte array
                byte[] byteArr = new byte[targetArray.Length * sizeof(ushort)];
                Buffer.BlockCopy(targetArray, 0, byteArr, 0, byteArr.Length);
                w.Write(byteArr);
                w.Close();
                fs.Close();
            }
        }
        #endregion Public Methods

        //**************************************************************
        #region Protected Methods

        private WriteFileSettingsClass GetWriteFile()
        {
            WriteFileSettingsClass writeFileClass = null;
            lock (ImageWriteLock)
            {
                if (WriteFileList.Count > 0)
                {
                    writeFileClass = WriteFileList[0];
                }
            }

            return writeFileClass;
        }

        private void InitializeWriteFiles()
        {
            ImageWriteLock = new object();
            WriteFileList = new List<WriteFileSettingsClass>();
        }

        private void InsertWriteFile(WriteFileSettingsClass writeFileClass)
        {
            lock (ImageWriteLock)
            {
                WriteFileList.Add(writeFileClass);
                if (WriteFileList.Count == 1)
                {
                    // Then we need to start the write thread
                    Thread _thread = new Thread(new ThreadStart(WriteFileThread));
                    _thread.Start();
                }
            }
        }

        private void PopWriteFile(WriteFileSettingsClass writeFileClass)
        {
            lock (ImageWriteLock)
            {
                WriteFileList.Remove(writeFileClass);
            }
        }

        private void WriteFileThread()
        {
            WriteFileSettingsClass writeFileClass = GetWriteFile();
            while (writeFileClass != null)
            {
                WriteFileFromClass(writeFileClass);
                writeFileClass.Frame.IsWritePending = false;
                PopWriteFile(writeFileClass);
                writeFileClass = GetWriteFile();
            }
        }

        private void WriteFileFromClass(WriteFileSettingsClass writeFileClass)
        {
            string fileName = String.Empty;
            BitmapSource sourceLow = null;
            BitmapSource sourceHigh = null;
            BitmapSource sourceMerge = null;
            List<string> nameList = new List<string>();

            try
            {

                if (writeFileClass.IsSaveExposureToFITFileChecked)
                {
                    if (writeFileClass.UseLowGainImageChecked)
                    {
                        fileName = GetValidFilename(writeFileClass.Folder, writeFileClass.FileNameBase + "Low_", FILE_EXTENSION_FIT);
                        nameList.Add(fileName);
                        FITSImageProcessing.SaveFitsFile(fileName, writeFileClass.Frame, FliFrame.FliImageType.LowGainDummy);
                    }
                    if (writeFileClass.UseHighGainImageChecked)
                    {
                        fileName = GetValidFilename(writeFileClass.Folder, writeFileClass.FileNameBase + "High_", FILE_EXTENSION_FIT);
                        nameList.Add(fileName);
                        FITSImageProcessing.SaveFitsFile(fileName, writeFileClass.Frame, FliFrame.FliImageType.HighGainDummy);
                    }
                    if (writeFileClass.UseMergedImageChecked)
                    {
                        fileName = GetValidFilename(writeFileClass.Folder, writeFileClass.FileNameBase + "Merge_", FILE_EXTENSION_FIT);
                        nameList.Add(fileName);
                        FITSImageProcessing.SaveFitsFile(fileName, writeFileClass.Frame, FliFrame.FliImageType.MergedDummy);
                    }
                }

                if (writeFileClass.IsSaveExposureToJPEGFileChecked)
                {
                    if (writeFileClass.UseLowGainImageChecked)
                    {
                        fileName = GetValidFilename(writeFileClass.Folder, writeFileClass.FileNameBase + "Low_", FILE_EXTENSION_JPEG);
                        nameList.Add(fileName);
                        sourceLow = SaveImageFile8Bit(writeFileClass.Frame, fileName, ImageFormat.Jpeg, sourceLow,
                            writeFileClass.Frame.GetTransformedPixelArray(FliFrame.FliImageType.LowGain));
                    }
                    if (writeFileClass.UseHighGainImageChecked)
                    {
                        fileName = GetValidFilename(writeFileClass.Folder, writeFileClass.FileNameBase + "High_", FILE_EXTENSION_JPEG);
                        sourceHigh = SaveImageFile8Bit(writeFileClass.Frame, fileName, ImageFormat.Jpeg, sourceHigh,
                            writeFileClass.Frame.GetTransformedPixelArray(FliFrame.FliImageType.HighGain));
                    }
                    if (writeFileClass.UseMergedImageChecked)
                    {
                        fileName = GetValidFilename(writeFileClass.Folder, writeFileClass.FileNameBase + "Merge_", FILE_EXTENSION_JPEG);
                        sourceMerge = SaveImageFile8Bit(writeFileClass.Frame, fileName, ImageFormat.Jpeg, sourceMerge,
                            writeFileClass.Frame.GetTransformedPixelArray(FliFrame.FliImageType.Merged));
                    }
                }

                if (writeFileClass.IsSaveExposureToPNGFileChecked)
                {
                    if (writeFileClass.UseLowGainImageChecked)
                    {
                        fileName = GetValidFilename(writeFileClass.Folder, writeFileClass.FileNameBase + "Low_", FILE_EXTENSION_PNG);
                        nameList.Add(fileName);
                        sourceLow = SaveImageFile(writeFileClass.Frame, fileName, ImageFormat.Png, sourceLow,
                            writeFileClass.Frame.GetPixelArray(FliFrame.FliImageType.LowGain));
                    }
                    if (writeFileClass.UseHighGainImageChecked)
                    {
                        fileName = GetValidFilename(writeFileClass.Folder, writeFileClass.FileNameBase + "High_", FILE_EXTENSION_PNG);
                        sourceHigh = SaveImageFile(writeFileClass.Frame, fileName, ImageFormat.Png, sourceHigh, writeFileClass.Frame.GetPixelArray(FliFrame.FliImageType.HighGain));
                    }
                    if (writeFileClass.UseMergedImageChecked)
                    {
                        fileName = GetValidFilename(writeFileClass.Folder, writeFileClass.FileNameBase + "Merge_", FILE_EXTENSION_PNG);
                        sourceMerge = SaveImageFile(writeFileClass.Frame, fileName, ImageFormat.Png, sourceMerge,
                            writeFileClass.Frame.GetPixelArray(FliFrame.FliImageType.Merged));
                    }
                }

                if (writeFileClass.IsSaveExposureToBINFileChecked)
                {
                    // First save the low only as a 16 bit file 
                    if (writeFileClass.UseLowGainImageChecked)
                    {
                        fileName = GetValidFilename(writeFileClass.Folder, writeFileClass.FileNameBase + "Low_", FILE_EXTENSION_BIN);
                        nameList.Add(fileName);
                        SaveBINFile(writeFileClass.Frame, fileName, File_Type.Low);
                    }
                    if (writeFileClass.UseHighGainImageChecked)
                    {
                        fileName = GetValidFilename(writeFileClass.Folder, writeFileClass.FileNameBase + "High_", FILE_EXTENSION_BIN);
                        SaveBINFile(writeFileClass.Frame, fileName, File_Type.High);
                    }
                    if (writeFileClass.UseMergedImageChecked)
                    {
                        fileName = GetValidFilename(writeFileClass.Folder, writeFileClass.FileNameBase + "Merge_", FILE_EXTENSION_BIN);
                        SaveBINFile(writeFileClass.Frame, fileName, File_Type.Merged);
                    }
                }

                if (writeFileClass.IsSaveExposureToRCDFileChecked || !String.IsNullOrEmpty(writeFileClass.NameHint))
                // We always want to write the final image from Test routines
                {
                    // Save the Combined file which is exactly how we received it from the board.
                    fileName = GetValidFilename(writeFileClass.Folder, writeFileClass.FileNameBase + "RCDData_", FILE_EXTENSION_RCD);
                    nameList.Add(fileName);
                    SaveBINFile(writeFileClass.Frame, fileName, File_Type.RawData);
                }

                if (writeFileClass.IsSaveExposureToRAWFileChecked)
                {
                    if (writeFileClass.UseLowGainImageChecked)
                    {
                        fileName = GetValidFilename(writeFileClass.Folder, writeFileClass.FileNameBase + "Low_", FILE_EXTENSION_RAW);
                        nameList.Add(fileName);
                        SaveRAWFile(writeFileClass.Frame.GetPixelArray(FliFrame.FliImageType.LowGain), fileName);
                    }
                    if (writeFileClass.UseHighGainImageChecked)
                    {
                        fileName = GetValidFilename(writeFileClass.Folder, writeFileClass.FileNameBase + "High_", FILE_EXTENSION_RAW);
                        SaveRAWFile(writeFileClass.Frame.GetPixelArray(FliFrame.FliImageType.HighGain), fileName);
                    }
                    if (writeFileClass.UseMergedImageChecked)
                    {
                        fileName = GetValidFilename(writeFileClass.Folder, writeFileClass.FileNameBase + "Merge_", FILE_EXTENSION_RAW);
                        SaveRAWFile(writeFileClass.Frame.GetPixelArray(FliFrame.FliImageType.Merged), fileName);
                    }
                }

                if (writeFileClass.IsSaveExposureToTIFFFileChecked)
                {
                    if (writeFileClass.UseLowGainImageChecked)
                    {
                        fileName = GetValidFilename(writeFileClass.Folder, writeFileClass.FileNameBase + "Low_", FILE_EXTENSION_TIFF);
                        nameList.Add(fileName);
                        sourceLow = SaveImageFile(writeFileClass.Frame, fileName, ImageFormat.Tiff, sourceLow,
                            writeFileClass.Frame.GetPixelArray(FliFrame.FliImageType.LowGain));
                    }
                    if (writeFileClass.UseHighGainImageChecked)
                    {
                        fileName = GetValidFilename(writeFileClass.Folder, writeFileClass.FileNameBase + "High_", FILE_EXTENSION_TIFF);
                        sourceHigh = SaveImageFile(writeFileClass.Frame, fileName, ImageFormat.Tiff, sourceHigh,
                            writeFileClass.Frame.GetPixelArray(FliFrame.FliImageType.HighGain));
                    }
                    if (writeFileClass.UseMergedImageChecked)
                    {
                        fileName = GetValidFilename(writeFileClass.Folder, writeFileClass.FileNameBase + "Merge_", FILE_EXTENSION_TIFF);
                        sourceMerge = SaveImageFile(writeFileClass.Frame, fileName, ImageFormat.Tiff, sourceMerge,
                            writeFileClass.Frame.GetPixelArray(FliFrame.FliImageType.Merged));
                    }
                }

                if (writeFileClass.IsSaveMetaDataToFileChecked)
                {
                    fileName = GetValidFilename(writeFileClass.Folder, writeFileClass.FileNameBase + "Metadata_", FILE_EXTENSION_TXT);
                    SaveMetadataFile(writeFileClass.Frame, fileName);
                }

                if (writeFileClass.SaveSectorsAsFiles)
                {
                    WriteFileSectors(writeFileClass.Frame, writeFileClass.Folder, nameList);
                }
            }
            catch (Exception ex)
            {
                if (!fliProCameraManager.FileWriteError)
                {
                    System.Windows.MessageBox.Show("Error encountered writing to folder:\n" + writeFileClass.Folder + "\n\n" + ex.Message);
                    fliProCameraManager.FileWriteError = true;
                }
            }
        }

        #endregion Protected Methods

    }
}
